import uuid

import lief

from chomper.types import Module


class IosModule:
    """System module on iOS."""

    MODULE_NAME: str

    def __init__(self, emu):
        self.emu = emu

        self.reloc_map = {}
        self.refs_reloc = set()

    def relocate_address(self, module_base: int, address: int) -> int:
        """Relocate the address."""
        if address in self.reloc_map:
            return self.reloc_map[address]

        value = self.emu.read_pointer(address)
        if not value:
            return value

        value += module_base

        self.emu.write_pointer(address, value)
        self.reloc_map[address] = value

        return value

    def set_refs_reloc(self, address: int):
        self.refs_reloc.add(address)

    def relocate_refs(self, binary, module_base: int):
        """Relocate all references."""
        for section in binary.sections:
            for pos in range(0, len(section.content) - 8 + 1):
                value = int.from_bytes(section.content[pos : pos + 8], "little")

                if value in self.refs_reloc:
                    address = module_base + section.virtual_address + pos
                    self.emu.write_pointer(address, module_base + value)

        self.refs_reloc.clear()

    def relocate_symbols(self, binary: lief.MachO.Binary):
        """Relocate references to all symbols."""
        for symbol in binary.symbols:
            if not symbol.value:
                continue

            self.set_refs_reloc(symbol.value)

    def relocate_block_invoke_calls(self, binary: lief.MachO.Binary, module_base: int):
        """Relocate function address for block struct."""
        for binding in binary.dyld_info.bindings:
            if binding.symbol.name == "__NSConcreteGlobalBlock":
                address = module_base + binding.address

                flags = self.emu.read_u32(address + 0x8)
                reversed_value = self.emu.read_u32(address + 0xC)

                if flags == 0x50000000 and reversed_value == 0:
                    self.relocate_address(module_base, address + 0x10)

    def fix_sections(self, binary: lief.MachO.Binary, module_base: int):
        """Fix sections which contains pointers."""
        section_names = [
            "__objc_classlist",
            "__objc_catlist",
            "__objc_nlcatlist",
            "__objc_protolist",
            "__objc_selrefs",
            "__objc_protorefs",
            "__la_resolver",
        ]

        for section_name in section_names:
            section = binary.get_section(section_name)
            if not section:
                continue

            begin = module_base + section.virtual_address
            end = begin + section.size

            values = self.emu.read_array(begin, end)

            if section_name == "__objc_classlist":
                for value in values:
                    self.fix_class_struct(module_base, value)
                    self.set_refs_reloc(value)

                    meta_class_ptr = self.emu.read_pointer(module_base + value)

                    self.fix_class_struct(module_base, meta_class_ptr)
                    self.set_refs_reloc(meta_class_ptr)

            elif section_name == "__objc_catlist":
                for value in values:
                    self.fix_category_struct(module_base, value)

            elif section_name == "__objc_protolist":
                for value in values:
                    self.fix_protocol_struct(module_base, value)

            elif section_name == "__objc_protorefs":
                for value in values:
                    self.relocate_address(module_base, module_base + value + 8)

            values = map(lambda v: module_base + v if v else 0, values)
            self.emu.write_array(begin, values)

        # Some system libraries don't contain `__objc_classlist`
        # or `__objc_protolist` section, so find struct by symbol name.
        if not binary.has_section("__objc_classlist"):
            for symbol in binary.symbols:
                symbol_name = str(symbol.name)

                if symbol.value and (
                    symbol_name.startswith("_OBJC_METACLASS_$")
                    or symbol_name.startswith("_OBJC_CLASS_$")
                ):
                    self.fix_class_struct(module_base, symbol.value)
                    self.set_refs_reloc(symbol.value)

        if not binary.has_section("__objc_protolist"):
            for symbol in binary.symbols:
                symbol_name = str(symbol.name)

                if symbol.value and symbol_name.startswith("__OBJC_PROTOCOL_$"):
                    self.fix_protocol_struct(module_base, symbol.value)

    def fix_method_list(self, module_base: int, address: int):
        """Fix the method list struct."""
        method_list_ptr = self.relocate_address(module_base, address)
        if not method_list_ptr:
            return

        flag = self.emu.read_u32(method_list_ptr)
        count = self.emu.read_u32(method_list_ptr + 4)

        # Set isFixedUp of method_list_t is False
        flag = flag & (~0x3)
        self.emu.write_u32(method_list_ptr, flag)

        if flag < 0x80000000:
            for i in range(count):
                method_ptr = method_list_ptr + 8 + i * 0x18

                # Fix method SEL
                self.relocate_address(module_base, method_ptr)

                # Fix method IMPL
                self.relocate_address(module_base, method_ptr + 0x10)

        else:
            for i in range(count):
                sel_offset_ptr = method_list_ptr + 8 + i * 12
                sel_offset = self.emu.read_u32(sel_offset_ptr)

                # Fix method SEL
                self.relocate_address(module_base, sel_offset_ptr + sel_offset)

    def fix_variable_list(self, module_base: int, address: int):
        """Fix the variable list struct."""
        variable_list_ptr = self.relocate_address(module_base, address)
        if not variable_list_ptr:
            return

        count = self.emu.read_u32(variable_list_ptr + 4)

        for i in range(count):
            variable_offset = variable_list_ptr + 8 + i * 0x18

            self.relocate_address(module_base, variable_offset)
            self.relocate_address(module_base, variable_offset + 0x8)
            self.relocate_address(module_base, variable_offset + 0x16)

    def fix_protocol_list(self, module_base: int, address: int):
        """Fix the protocol list struct."""
        protocol_list_ptr = self.relocate_address(module_base, address)
        if not protocol_list_ptr:
            return

        count = self.emu.read_u64(protocol_list_ptr)

        for i in range(count):
            protocol_offset = protocol_list_ptr + 8 + i * 0x8

            protocol_address = self.relocate_address(module_base, protocol_offset)
            self.relocate_address(module_base, protocol_address + 0x8)

    def fix_class_struct(self, module_base: int, address: int):
        """Fix the class struct."""
        address += module_base

        data_ptr = self.emu.read_pointer(address + 0x20)
        self.set_refs_reloc(data_ptr)

        data_ptr += module_base

        name_ptr = self.emu.read_pointer(data_ptr + 0x18)
        if name_ptr:
            self.set_refs_reloc(name_ptr)

        instance_methods_ptr = data_ptr + 0x20
        self.fix_method_list(module_base, instance_methods_ptr)

        protocols_ptr = data_ptr + 0x28
        self.fix_protocol_list(module_base, protocols_ptr)

        instance_variables_ptr = data_ptr + 0x30
        self.fix_variable_list(module_base, instance_variables_ptr)

    def fix_protocol_struct(self, module_base: int, address: int):
        """Fix the protocol struct."""
        address += module_base

        name_ptr = self.emu.read_pointer(address + 0x8)
        if name_ptr:
            self.set_refs_reloc(name_ptr)

        protocols_ptr = address + 0x10
        self.fix_protocol_list(module_base, protocols_ptr)

    def fix_category_struct(self, module_base: int, address: int):
        """Fix the category struct."""
        address += module_base

        instance_methods_ptr = address + 0x10
        self.fix_method_list(module_base, instance_methods_ptr)

        class_methods_ptr = address + 0x18
        self.fix_method_list(module_base, class_methods_ptr)

    def fix_cstring_section(self, binary: lief.MachO.Binary):
        """Fix the `__cstring` section."""
        section = binary.get_section("__cstring")
        if not section:
            return

        section_content = bytes(section.content)
        if not section_content:
            return

        start = 0

        for offset in range(section.size):
            if section_content[offset] == 0:
                self.set_refs_reloc(section.virtual_address + start)
                start = offset + 1

    def fix_cfstring_section(self, binary: lief.MachO.Binary, module_base: int):
        """Fix the `__cfstring` section."""
        section_map = {
            "__cfstring": 0x20,
            "__cfstring_CFN": 0x38,
        }

        for section_name, step in section_map.items():
            section = binary.get_section(section_name)
            if not section:
                continue

            start = section.virtual_address
            offset = 0

            while offset < section.size:
                address = module_base + start + offset + 0x10
                self.relocate_address(module_base, address)
                self.set_refs_reloc(start + offset)
                offset += step

    def fix_cf_uni_char_string(self, binary: lief.MachO.Binary):
        """Fix references of the string which used by `__CFUniCharLoadFile`."""
        section_names = ["__csbitmaps", "__data", "__properties"]

        for section_name in section_names:
            section = binary.get_section("__UNICODE", section_name)
            if section:
                self.set_refs_reloc(section.virtual_address)

    def fix_image(self, module: Module):
        """Fix the image."""
        if not module.binary:
            return

        module_base = module.base - (module.image_base or 0)

        self.relocate_symbols(module.binary)
        self.relocate_block_invoke_calls(module.binary, module_base)
        self.fix_sections(module.binary, module_base)
        self.fix_cstring_section(module.binary)
        self.fix_cfstring_section(module.binary, module_base)
        self.fix_cf_uni_char_string(module.binary)
        self.relocate_refs(module.binary, module_base)

    def init_objc(self, module: Module):
        self.emu.os.init_objc(module)

    def initialize(self, module: Module):
        self.fix_image(module)
        self.init_objc(module)

    def load(self) -> Module:
        module_path = self.emu.os.find_system_module(self.MODULE_NAME)
        module = self.emu.load_module(module_path, exec_objc_init=False)

        self.initialize(module)

        return module


class SystemPlatform(IosModule):
    MODULE_NAME = "libsystem_platform.dylib"

    def set_arch_flag(self):
        """Set a flag meaning the arch type, which will be read by functions
        such as `_os_unfair_recursive_lock_lock_with_options`."""
        self.emu.uc.mem_map(0xFFFFFC000, 1024)

        # arch flag
        self.emu.write_u64(0xFFFFFC023, 2)

        self.emu.write_u64(0xFFFFFC104, 0x100)

    def initialize(self, module):
        self.set_arch_flag()
        super().initialize(module)


class SystemKernel(IosModule):
    MODULE_NAME = "libsystem_kernel.dylib"


class SystemC(IosModule):
    MODULE_NAME = "libsystem_c.dylib"

    def init_program_vars(self):
        """Initialize program variables."""
        argc = self.emu.create_buffer(8)
        self.emu.write_int(argc, 0, 8)

        nxargc_pointer = self.emu.find_symbol("_NXArgc_pointer")
        self.emu.write_pointer(nxargc_pointer.address, argc)

        nxargv_pointer = self.emu.find_symbol("_NXArgv_pointer")
        self.emu.write_pointer(nxargv_pointer.address, self.emu.create_string(""))

        environ_dict = {
            "__CF_USER_TEXT_ENCODING": "0:0",
            "CFN_USE_HTTP3": "0",
            "CFStringDisableROM": "1",
            "HOME": (
                f"/Users/Sergey/Library/Developer/CoreSimulator/Devices/{uuid.uuid4()}"
                f"/data/Containers/Data/Application/{uuid.uuid4()}"
            ),
        }

        size = self.emu.arch.addr_size * len(environ_dict) + 1
        environ_values = self.emu.create_buffer(size)

        offset = 0x0

        for key, value in environ_dict.items():
            prop_str = self.emu.create_string(f"{key}={value}")
            self.emu.write_pointer(environ_values + offset, prop_str)
            offset += self.emu.arch.addr_size

        self.emu.write_pointer(environ_values + offset, 0)

        environ = self.emu.create_buffer(8)
        self.emu.write_pointer(environ, environ_values)

        environ_pointer = self.emu.find_symbol("_environ_pointer")
        self.emu.write_pointer(environ_pointer.address, environ)

        progname_pointer = self.emu.find_symbol("___progname_pointer")
        self.emu.write_pointer(progname_pointer.address, self.emu.create_string(""))

    def initialize(self, module):
        super().initialize(module)

        symbol_names = ("___locale_key", "___global_locale")

        for symbol_name in symbol_names:
            symbol = self.emu.find_symbol(symbol_name)
            self.emu.write_pointer(symbol.address, 0)

        self.init_program_vars()


class SystemPthread(IosModule):
    MODULE_NAME = "libsystem_pthread.dylib"


class SystemInfo(IosModule):
    MODULE_NAME = "libsystem_info.dylib"


class SystemDarwin(IosModule):
    MODULE_NAME = "libsystem_darwin.dylib"


class SystemFeatureflags(IosModule):
    MODULE_NAME = "libsystem_featureflags.dylib"


class Corecrypto(IosModule):
    MODULE_NAME = "libcorecrypto.dylib"


class CommonCrypto(IosModule):
    MODULE_NAME = "libcommonCrypto.dylib"


class Cppabi(IosModule):
    MODULE_NAME = "libc++abi.dylib"


class Cpp(IosModule):
    MODULE_NAME = "libc++.1.dylib"


class Dyld(IosModule):
    MODULE_NAME = "libdyld.dylib"

    def initialize(self, module):
        super().initialize(module)

        g_use_dyld3 = self.emu.find_symbol("_gUseDyld3")
        self.emu.write_u8(g_use_dyld3.address, 1)

        dyld_all_images = self.emu.find_symbol("__ZN5dyld310gAllImagesE")

        # dyld3::closure::ContainerTypedBytes::findAttributePayload
        attribute_payload_ptr = self.emu.create_buffer(8)

        self.emu.write_u32(attribute_payload_ptr, 2**10)
        self.emu.write_u8(attribute_payload_ptr + 4, 0x20)

        self.emu.write_pointer(dyld_all_images.address, attribute_payload_ptr)

        # dyld3::AllImages::platform
        platform_ptr = self.emu.create_buffer(0x144)
        self.emu.write_u32(platform_ptr + 0x140, 2)

        self.emu.write_pointer(dyld_all_images.address + 0x50, platform_ptr)


class Objc(IosModule):
    MODULE_NAME = "libobjc.A.dylib"

    def initialize(self, module):
        self.fix_image(module)

        prototypes = self.emu.find_symbol("__ZL10prototypes")
        self.emu.write_u64(prototypes.address, 0)

        gdb_objc_realized_classes = self.emu.find_symbol("_gdb_objc_realized_classes")
        protocolsv_ret = self.emu.call_symbol("__ZL9protocolsv")

        self.emu.write_pointer(gdb_objc_realized_classes.address, protocolsv_ret)

        opt = self.emu.find_symbol("__ZL3opt")
        self.emu.write_pointer(opt.address, 0)

        # Disable pre-optimization
        disable_preopt = self.emu.find_symbol("_DisablePreopt")
        self.emu.write_u8(disable_preopt.address, 1)

        self.emu.call_symbol("__objc_init")

        self.init_objc(module)


class Dispatch(IosModule):
    MODULE_NAME = "libdispatch.dylib"

    def initialize(self, module):
        symbol = self.emu.find_symbol("_OBJC_CLASS_$_OS_dispatch_queue_serial")
        offsets = [0x30, 0x50, 0x58]

        for offset in offsets:
            address = self.emu.read_pointer(symbol.address + offset)
            self.set_refs_reloc(address)

        super().initialize(module)


class SystemBlocks(IosModule):
    MODULE_NAME = "libsystem_blocks.dylib"


class SystemTrace(IosModule):
    MODULE_NAME = "libsystem_trace.dylib"

    def initialize(self, module):
        super().initialize(module)

        self.emu.write_u64(0xFFFFFC104, 0x100)


class SystemSandbox(IosModule):
    MODULE_NAME = "libsystem_sandbox.dylib"


class Network(IosModule):
    MODULE_NAME = "libnetwork.dylib"


class CoreFoundation(IosModule):
    MODULE_NAME = "CoreFoundation"

    def initialize(self, module):
        super().initialize(module)

        cf_process_path_str = (
            f"/private/var/containers/Bundle/Application/{uuid.uuid4()}/App.app"
        )
        cf_process_path_ptr = self.emu.create_string(cf_process_path_str)

        cf_process_path = self.emu.find_symbol("___CFProcessPath")
        self.emu.write_pointer(cf_process_path.address, cf_process_path_ptr)

        self.emu.call_symbol("___CFInitialize")


class CFNetwork(IosModule):
    MODULE_NAME = "CFNetwork"

    def initialize(self, module):
        offset = module.base - module.image_base + 0x1D1C28610

        for i in range(25):
            address = self.emu.read_pointer(offset + i * 8)
            self.set_refs_reloc(address)

        self.set_refs_reloc(0x180A63880)

        super().initialize(module)


class Foundation(IosModule):
    MODULE_NAME = "Foundation"

    def initialize(self, module):
        super().initialize(module)

        self.emu.call_symbol("__NSInitializePlatform")


class Security(IosModule):
    MODULE_NAME = "Security"


class QuartzCore(IosModule):
    MODULE_NAME = "QuartzCore"


class BaseBoard(IosModule):
    MODULE_NAME = "BaseBoard"


class FrontBoardServices(IosModule):
    MODULE_NAME = "FrontBoardServices"


class PrototypeTools(IosModule):
    MODULE_NAME = "PrototypeTools"


class TextInput(IosModule):
    MODULE_NAME = "TextInput"


class PhysicsKit(IosModule):
    MODULE_NAME = "PhysicsKit"


class CoreAutoLayout(IosModule):
    MODULE_NAME = "CoreAutoLayout"


class UIFoundation(IosModule):
    MODULE_NAME = "UIFoundation"


class UIKitServices(IosModule):
    MODULE_NAME = "UIKitServices"


class UIKitCore(IosModule):
    MODULE_NAME = "UIKitCore"



import os

import pytest

from . import conftest

from chomper import Chomper
from chomper.const import ARCH_ARM64
from chomper.exceptions import EmulatorCrashedException


class TestExceptions:
    def test_unhandled_system_call_exception(self):
        with pytest.raises(EmulatorCrashedException, match=r"Unhandled system call.*"):
            emu = Chomper(arch=ARCH_ARM64)
            emu.hooks.pop("malloc")

            emu.load_module(os.path.join(conftest.android_lib64_path, "libc.so"))

            emu.call_symbol("malloc")

    def test_missing_symbol_required_exception(self, bytes_test):
        with pytest.raises(EmulatorCrashedException, match=r"Missing symbol.*"):
            emu = Chomper(arch=ARCH_ARM64)
            libszstone = emu.load_module(
                os.path.join(conftest.com_shizhuang_duapp_path, "libszstone.so")
            )

            a1 = emu.create_buffer(len(bytes_test))
            a2 = len(bytes_test)
            a3 = emu.create_buffer(1024)

            emu.write_bytes(a1, bytes_test)
            emu.call_address(libszstone.base + 0x289A4, a1, a2, a3)



import logging
import re
from functools import wraps
from typing import Callable, Dict, List, Optional, Tuple, Type, Union

from capstone import CS_ARCH_ARM, CS_ARCH_ARM64, CS_MODE_ARM, CS_MODE_THUMB, Cs
from unicorn import (
    UC_ARCH_ARM,
    UC_ARCH_ARM64,
    UC_HOOK_CODE,
    UC_HOOK_INTR,
    UC_MODE_ARM,
    UC_MODE_THUMB,
    Uc,
    UcError,
    arm64_const,
    arm_const,
)
from unicorn.unicorn import UC_HOOK_CODE_TYPE

from . import const
from .arch import arm_arch, arm64_arch
from .exceptions import EmulatorCrashedException, SymbolMissingException
from .memory import MemoryManager
from .types import Module, Symbol, OptionsTV
from .log import get_logger
from .os.android import AndroidOs
from .os.ios import IosOs
from .utils import aligned


class Chomper:
    """Lightweight emulation framework for mobile platform (Android, iOS).

    Args:
        arch: The architecture to emulate, support ARM and ARM64.
        mode: The emulation mode.
        os_type: The os type, support Android and iOS.
        logger: The logger to print log.
        endian: Default endian to use.
        enable_vfp: Enable VFP extension of ARM.
        trace_inst: Print log when any instruction is executed. The emulator will
            call disassembler in real time to display the assembly instructions,
            so this will slow down the emulation.
        trace_symbol_calls: Print log when any symbol is called.
    """

    def __init__(
        self,
        arch: int = const.ARCH_ARM64,
        mode: int = const.MODE_ARM,
        os_type: int = const.OS_ANDROID,
        logger: Optional[logging.Logger] = None,
        endian: const.EndianType = const.LITTLE_ENDIAN,
        rootfs_path: Optional[str] = None,
        enable_vfp: bool = True,
        trace_inst: bool = False,
        trace_symbol_calls: bool = False,
        os_options: Optional[OptionsTV] = None,
    ):
        self._setup_arch(arch)

        self.uc = self._create_uc(arch, mode)
        self.cs = self._create_cs(arch, mode)

        self.logger = logger or get_logger(self.__class__.__name__)

        self.os_type = os_type
        self.endian = endian

        self._trace_inst = trace_inst
        self._trace_symbol_calls = trace_symbol_calls

        self.modules: List[Module] = []

        self.hooks: Dict[str, UC_HOOK_CODE_TYPE] = {}
        self.syscall_handlers: Dict[int, Callable] = {}

        self.memory_manager = MemoryManager(
            uc=self.uc,
            address=const.HEAP_ADDRESS,
            minimum_pool_size=const.MINIMUM_POOL_SIZE,
        )

        self._setup_emulator(enable_vfp=enable_vfp)
        self._setup_interrupt_handler()

        os_cls: Union[Type[AndroidOs], Type[IosOs]]

        if os_type == const.OS_ANDROID:
            os_cls = AndroidOs
        elif os_type == const.OS_IOS:
            os_cls = IosOs
        else:
            raise ValueError("Unsupported platform type")

        self.os = os_cls(self, rootfs_path=rootfs_path, options=os_options)
        self.os.initialize()

    def _setup_arch(self, arch: int):
        """Setup architecture."""
        if arch == const.ARCH_ARM:
            self.arch = arm_arch
        elif arch == const.ARCH_ARM64:
            self.arch = arm64_arch
        else:
            raise ValueError("Invalid argument arch")

    @staticmethod
    def _create_uc(arch: int, mode: int) -> Uc:
        """Create Unicorn instance."""
        arch = UC_ARCH_ARM if arch == const.ARCH_ARM else UC_ARCH_ARM64
        mode = UC_MODE_THUMB if mode == const.MODE_THUMB else UC_MODE_ARM

        return Uc(arch, mode)

    @staticmethod
    def _create_cs(arch: int, mode: int) -> Cs:
        """Create Capstone instance."""
        arch = CS_ARCH_ARM if arch == const.ARCH_ARM else CS_ARCH_ARM64
        mode = CS_MODE_THUMB if mode == const.MODE_THUMB else CS_MODE_ARM

        return Cs(arch, mode)

    def _setup_stack(self):
        """Setup stack."""
        stack_addr = const.STACK_ADDRESS + const.STACK_SIZE // 2
        self.uc.mem_map(const.STACK_ADDRESS, const.STACK_SIZE)

        self.uc.reg_write(self.arch.reg_sp, stack_addr)
        self.uc.reg_write(self.arch.reg_fp, stack_addr)

    def _setup_thread_register(self):
        """Setup thread register.

        The thread register store the address of thread local storage (TLS).

        The function only allocates a block of memory to TLS and doesn't really
        initialize.
        """
        self.uc.mem_map(const.TLS_ADDRESS - 1024, const.TLS_SIZE + 1024)

        if self.os_type == const.OS_IOS:
            self.uc.reg_write(arm64_const.UC_ARM64_REG_TPIDRRO_EL0, const.TLS_ADDRESS)

        else:
            if self.arch == arm_arch:
                self.uc.reg_write(
                    arm_const.UC_ARM_REG_CP_REG,
                    (15, 0, 0, 13, 0, 0, 3, const.TLS_ADDRESS),  # type: ignore
                )

            elif self.arch == arm64_arch:
                self.uc.reg_write(arm64_const.UC_ARM64_REG_TPIDR_EL0, const.TLS_ADDRESS)

    def _enable_vfp(self):
        """Enable vfp.

        See details:
        https://github.com/unicorn-engine/unicorn/issues/446
        """
        inst_code = (
            b"\x4f\xf4\x70\x00"  # mov.w r0, #0xf00000
            b"\x01\xee\x50\x0f"  # mcr p15, #0x0, r0, c1, c0, #0x2
            b"\xbf\xf3\x6f\x8f"  # isb sy
            b"\x4f\xf0\x80\x43"  # mov.w r3, #0x40000000
            b"\xe8\xee\x10\x3a"  # vmsr fpexc, r3
        )
        addr = self.create_buffer(1024)

        self.uc.mem_write(addr, inst_code)
        self.uc.emu_start(addr | 1, addr + len(inst_code))

        self.free(addr)

    def _setup_emulator(self, enable_vfp: bool = True):
        """Setup emulator."""
        self._setup_stack()
        self._setup_thread_register()

        if self.arch == arm_arch and enable_vfp:
            self._enable_vfp()

    def _start_emulate(self, address: int, *args: int):
        """Start emulate at the specified address."""
        stop_addr = self.create_buffer(8)
        # stop_addr = 0

        for index, value in enumerate(args):
            self.set_arg(index, value)

        # Set the value of register LR to the stop address of the emulation,
        # so that when the function returns, it will jump to this address.
        self.uc.reg_write(self.arch.reg_lr, stop_addr)
        self.uc.reg_write(self.arch.reg_pc, address)
        self.uc.reg_write(self.arch.reg_fp, 0)

        try:
            self.logger.info(f"Start emulate at {self.debug_symbol(address)}")
            self.uc.emu_start(address, stop_addr)

        except UcError as e:
            self.crash("Unknown reason", from_exc=e)

        finally:
            self.free(stop_addr)

    def find_module(self, name_or_addr: Union[str, int]) -> Optional[Module]:
        """Find module by name or address."""
        for module in self.modules:
            if isinstance(name_or_addr, str):
                if module.name == name_or_addr:
                    return module

            elif isinstance(name_or_addr, int):
                if module.base <= name_or_addr < module.base + module.size:
                    return module

        return None

    def find_symbol(self, symbol_name: str) -> Symbol:
        """Find symbol from loaded modules.

        Raises:
            SymbolMissingException: If symbol not found.
        """
        for module in self.modules:
            for symbol in module.symbols:
                if symbol.name == symbol_name:
                    return symbol

        raise SymbolMissingException(f"{symbol_name} not found")

    def debug_symbol(self, address: int) -> str:
        """Format address to `libtest.so!0x1000` or `0x10000`."""
        module = self.find_module(address)

        if module:
            offset = address - module.base + (module.image_base or 0)
            return f"{module.name}!0x{offset:x}"

        return f"0x{address:x}"

    def backtrace(self) -> List[Tuple[int, Optional[Module]]]:
        """Backtrace call stack."""
        stack = [self.uc.reg_read(self.arch.reg_pc)]
        stack += [self.uc.reg_read(self.arch.reg_lr) - 4]

        frame = self.uc.reg_read(self.arch.reg_fp)
        limit = 32

        for _ in range(limit - 1):
            address = self.read_pointer(frame + 8)
            frame = self.read_pointer(frame)

            if not frame or not address:
                break

            if address < const.MODULE_ADDRESS:
                continue

            stack.append(address - 4)

        return [(address, self.find_module(address)) for address in stack if address]

    def add_hook(
        self,
        symbol_or_addr: Union[int, str],
        callback: UC_HOOK_CODE_TYPE,
        user_data: Optional[dict] = None,
    ) -> int:
        """Add hook to the emulator.

        Args:
            symbol_or_addr: The symbol name or the address to hook. If this is ``str``,
                the function will look up symbol from loaded modules and use its
                address to hook.
            callback: The callback function, same as callback of type ``UC_HOOK_CODE``
                in unicorn.
            user_data: A ``dict`` that contains the data you want to pass to the
                callback function. The ``Chomper`` instance will also be passed in
                as an emulator field.

        Raises:
            SymbolMissingException: If symbol not found.
        """
        if isinstance(symbol_or_addr, int):
            hook_addr = symbol_or_addr

        else:
            symbol = self.find_symbol(symbol_or_addr)
            hook_addr = symbol.address

        if self.arch == arm_arch:
            hook_addr = (hook_addr | 1) - 1

        return self.uc.hook_add(
            UC_HOOK_CODE,
            callback,
            begin=hook_addr,
            end=hook_addr,
            user_data={"emu": self, **(user_data or {})},
        )

    def add_interceptor(
        self,
        symbol_or_addr: Union[int, str],
        callback: UC_HOOK_CODE_TYPE,
        user_data: Optional[dict] = None,
    ):
        """Add interceptor to the emulator."""

        @wraps(callback)
        def decorator(uc, address, size, user_data_):
            emu = user_data_["emu"]
            address = emu.uc.reg_read(emu.arch.reg_pc)

            retval = callback(uc, address, size, user_data_)

            if isinstance(retval, int):
                emu.set_retval(retval)

            if address == emu.uc.reg_read(emu.arch.reg_pc):
                emu.uc.reg_write(emu.arch.reg_pc, emu.uc.reg_read(emu.arch.reg_lr))

        return self.add_hook(symbol_or_addr, decorator, user_data)

    def del_hook(self, handle: int):
        """Delete hook."""
        self.uc.hook_del(handle)

    def log_registers(self):
        """Log all register values."""
        message = "Registers: "

        for reg_index in range(31):
            reg_id = getattr(arm64_const, f"UC_ARM64_REG_X{reg_index}")
            reg_value = self.uc.reg_read(reg_id)

            if reg_index:
                message += ", "

            message += f"x{reg_index}: 0x{reg_value:016x}"

            if self.find_module(reg_value):
                message += f" [{self.debug_symbol(reg_value)}]"

        self.logger.info(message)

    def log_trace(self):
        """Log the trace stack."""
        trace_stack = [self.debug_symbol(t[0]) for t in self.backtrace()]
        message = "Trace stack: %s" % ", ".join(trace_stack)
        self.logger.info(message)

    def crash(self, message: str, from_exc: Optional[Exception] = None):
        """Raise an emulator crashed exception and output debugging info.

        Raises:
            EmulatorCrashedException:
        """
        self.log_registers()
        self.log_trace()

        address = self.uc.reg_read(self.arch.reg_pc)
        message = f"{message} at {self.debug_symbol(address)}"

        if from_exc:
            raise EmulatorCrashedException(message) from from_exc

        raise EmulatorCrashedException(message)

    def trace_symbol_call_callback(self, *args):
        """Trace symbol call."""
        user_data = args[-1]
        symbol = user_data["symbol"]

        if self.arch == const.ARCH_ARM:
            ret_addr = self.uc.reg_read(arm_const.UC_ARM_REG_LR)
        else:
            ret_addr = self.uc.reg_read(arm64_const.UC_ARM64_REG_LR)

        self.logger.info(
            f'Symbol "{symbol.name}" called'
            + (f" from {self.debug_symbol(ret_addr)}" if ret_addr else "")
        )

    def trace_inst_callback(self, uc: Uc, address: int, size: int, user_data: dict):
        """Trace instruction."""
        for inst in self.cs.disasm_lite(uc.mem_read(address, size), 0):
            self.logger.info(
                f"Trace at {self.debug_symbol(address)}: {inst[-2]} {inst[-1]}"
            )

    @staticmethod
    def missing_symbol_required_callback(*args):
        """Raise an exception with information of missing symbol."""
        user_data = args[-1]
        symbol_name = user_data["symbol_name"]

        raise EmulatorCrashedException(
            f'Missing symbol "{symbol_name}" is required, '
            f"you should load the library that contains it."
        )

    def _setup_interrupt_handler(self):
        """Setup interrupt handler."""
        self.uc.hook_add(UC_HOOK_INTR, self._interrupt_callback)

    def _interrupt_callback(self, uc: Uc, intno: int, user_data: dict):
        """Handle interrupt from emulators.

        There are currently two types of interrupts that need to be actively handled,
        system calls and some unsupported instructions.
        """
        if intno == 2:
            self._dispatch_syscall()
            return

        elif intno in (1, 4):
            address = self.uc.reg_read(self.arch.reg_pc)
            inst = next(self.cs.disasm_lite(uc.mem_read(address, 4), 0))
            result = self._handle_atomic_inst(inst)

            if result:
                self.uc.reg_write(arm64_const.UC_ARM64_REG_PC, address + 4)
                return

        self.crash("Unhandled interruption %s" % (intno,))

    def _dispatch_syscall(self):
        """Dispatch system calls to the registered handlers of the OS."""
        if self.os_type == const.OS_IOS:
            syscall_id = self.uc.reg_read(arm64_const.UC_ARM64_REG_X16)
            syscall_handler = self.syscall_handlers.get(syscall_id)

            if syscall_handler:
                syscall_handler(self)
                return

        self.crash("Unhandled system call")

    def _handle_atomic_inst(self, inst: Tuple[int, int, str, str]) -> bool:
        """Handle atomic instructions.

        The iOS system libraries will use some atomic instructions from ARM v8.1.
        However, Unicorn doesn't support these instructions, so we need to do some
        simple simulation.

        Returns:
            True if the instruction is handled, False otherwise.
        """
        inst_set = ["ldxr", "ldadd", "ldset", "swp", "cas"]

        if not any((inst[2].startswith(t) for t in inst_set)):
            return False

        match = re.match(r"(\w+), \[(\w+)]", inst[3])

        if not match:
            match = re.match(r"(\w+), (\w+), \[(\w+)]", inst[3])

        if not match:
            return False

        if inst[2].endswith("b"):
            op_bits = 8
        elif re.search(r"w(\d+)", inst[3]):
            op_bits = 32
        else:
            op_bits = 64

        regs = []

        for reg in match.groups():
            attr = f"UC_ARM64_REG_{reg.upper()}"
            regs.append(getattr(arm64_const, attr))

        return self._exec_atomic_inst(inst[2], op_bits, regs)

    def _exec_atomic_inst(self, inst: str, op_bits: int, regs: List[int]) -> bool:
        """Execute an atomic instruction."""
        exec_func_map = {
            "ldxr": self._exec_inst_ldxr,
            "ldadd": self._exec_inst_ldadd,
            "ldset": self._exec_inst_ldset,
            "swp": self._exec_inst_swp,
            "cas": self._exec_inst_cas,
        }

        read_func = getattr(self, f"read_u{op_bits}")
        write_func = getattr(self, f"write_u{op_bits}")

        addr = self.uc.reg_read(regs[-1])
        value = read_func(addr)

        for key, exec_func in exec_func_map.items():
            if inst.startswith(key):
                value = exec_func(regs, value)

                if value is not None:
                    write_func(addr, value % (2**op_bits))

                return True

        return False

    def _exec_inst_ldxr(self, regs: List[int], value: int):
        """Execute `ldxr` instruction."""
        self.uc.reg_write(regs[0], value)
        return None

    def _exec_inst_ldadd(self, regs: List[int], value: int):
        """Execute `ldadd` instruction."""
        self.uc.reg_write(regs[1], value)
        value += self.uc.reg_read(regs[0])
        return value

    def _exec_inst_ldset(self, regs: List[int], value: int):
        """Execute `ldset` instruction."""
        self.uc.reg_write(regs[1], value)
        value |= self.uc.reg_read(regs[0])
        return value

    def _exec_inst_swp(self, regs: List[int], value: int):
        """Execute `swp` instruction."""
        self.uc.reg_write(regs[1], value)
        value = self.uc.reg_read(regs[0])
        return value

    def _exec_inst_cas(self, regs: List[int], value: int):
        """Execute `cas` instruction."""
        self.uc.reg_write(regs[0], value)

        if self.uc.reg_read(regs[0]) == value:
            value = self.uc.reg_read(regs[1])
            return value

        return None

    def add_inst_trace(self, module: Module):
        """Add instruction trace for the module."""
        self.uc.hook_add(
            UC_HOOK_CODE,
            self.trace_inst_callback,
            begin=module.base,
            end=module.base + module.size,
        )

    def exec_init_array(self, init_array: List[int]):
        """Execute initialization functions."""
        for init_func in init_array:
            if not init_func:
                continue

            try:
                self.logger.info(
                    "Execute initialization function %s" % self.debug_symbol(init_func)
                )
                self.call_address(init_func)

            except Exception as e:
                self.logger.warning(
                    "c %s execute failed: %s" % (self.debug_symbol(init_func), repr(e))
                )

    def load_module(
        self,
        module_file: str,
        exec_init_array: bool = False,
        exec_objc_init: bool = True,
        trace_inst: bool = False,
        trace_symbol_calls: bool = False,
    ) -> Module:
        """Load executable file from path.

        Args:
            module_file: The path of file to be loaded.
            exec_init_array: Execute initialization functions recorded in the section
                `.init_array` after the module loaded.
            exec_objc_init: Execute `_objc_init` function after the module loaded.
            trace_inst: Output log when the instructions in this module are executed.
                The emulator will call disassembler in real time to display the
                assembly instructions, so this will slow down the emulation.
            trace_symbol_calls: Output log when the symbols in this module are called.
        """
        if not self.modules:
            module_base = const.MODULE_ADDRESS
        else:
            prev = self.modules[-1]
            module_base = aligned(prev.base + prev.size, 1024 * 1024)

        module = self.os.loader.load(
            module_base=module_base,
            module_file=module_file,
            trace_symbol_calls=trace_symbol_calls or self._trace_symbol_calls,
        )
        self.modules.append(module)

        # Trace instructions
        if trace_inst or self._trace_inst:
            self.add_inst_trace(module)

        if exec_objc_init and isinstance(self.os, IosOs):
            self.os.init_objc(module)

        if exec_init_array:
            self.exec_init_array(module.init_array)

        return module

    def _get_arg_container(self, index: int) -> Tuple[bool, int]:
        """Get the register or address where the specified argument is stored.

        On arch ARM, the first four parameters are stored in the register R0-R3, and
        the rest are stored on the stack. On arch ARM64, the first eight parameters
        are stored in the register X0-X7, and the rest are stored on the stack.

        Returns:
            A ``tuple`` contains a ``bool`` and an ``int``, the first member means
            whether the returned is a register and the second is the actual register
            or address.
        """
        if index >= len(self.arch.reg_args):
            # Read address from stack.
            offset = (index - len(self.arch.reg_args)) * self.arch.addr_size
            address = self.uc.reg_read(self.arch.reg_sp) + offset
            return False, address

        return True, self.arch.reg_args[index]

    def get_arg(self, index: int) -> int:
        """Get argument with the specified index."""
        is_reg, reg_or_addr = self._get_arg_container(index)

        if is_reg:
            return self.uc.reg_read(reg_or_addr)
        else:
            return self.read_int(reg_or_addr, self.arch.addr_size)

    def set_arg(self, index: int, value: int):
        """Set argument with the specified index."""
        is_reg, reg_or_addr = self._get_arg_container(index)

        if is_reg:
            self.uc.reg_write(reg_or_addr, value)
        else:
            self.write_int(reg_or_addr, value, self.arch.addr_size)

    def get_retval(self) -> int:
        """Get return value."""
        return self.uc.reg_read(self.arch.reg_retval)

    def set_retval(self, value: int):
        """Set return value."""
        self.uc.reg_write(self.arch.reg_retval, value)

    def create_buffer(self, size: int) -> int:
        """Create a buffer with the size."""
        return self.memory_manager.alloc(size)

    def create_string(self, string: str) -> int:
        """Create a buffer that is initialized to the string."""
        address = self.memory_manager.alloc(len(string) + 1)
        self.write_string(address, string)

        return address

    def free(self, address: int):
        """Free memory."""
        self.memory_manager.free(address)

    def read_int(self, address: int, size: int, signed: bool = False) -> int:
        """Read an integer from the address."""
        return int.from_bytes(
            self.uc.mem_read(address, size),
            signed=signed,
            byteorder=self.endian,
        )

    def read_s8(self, address: int):
        """Read a signed int8 from the address."""
        return self.read_int(address, 1, True)

    def read_s16(self, address: int):
        """Read a signed int16 from the address."""
        return self.read_int(address, 2, True)

    def read_s32(self, address: int):
        """Read a signed int32 from the address."""
        return self.read_int(address, 4, True)

    def read_s64(self, address: int):
        """Read a signed int64 from the address."""
        return self.read_int(address, 8, True)

    def read_u8(self, address: int):
        """Read an unsigned int8 from the address."""
        return self.read_int(address, 1, False)

    def read_u16(self, address: int):
        """Read an unsigned int16 from the address."""
        return self.read_int(address, 2, False)

    def read_u32(self, address: int):
        """Read an unsigned int32 from the address."""
        return self.read_int(address, 4, False)

    def read_u64(self, address: int):
        """Read an unsigned int64 from the address."""
        return self.read_int(address, 8, False)

    def read_bytes(self, address: int, size: int) -> bytes:
        """Read bytes from the address."""
        return bytes(self.uc.mem_read(address, size))

    def read_string(self, address: int) -> str:
        """Read string from the address."""
        data = bytes()
        offset = 0

        while True:
            byte = self.read_bytes(address + offset, 1)
            if byte == b"\x00":
                break

            data += byte
            offset += 1

        return data.decode("utf-8")

    def read_pointer(self, address: int) -> int:
        """Read a pointer from the address."""
        return self.read_int(address, self.arch.addr_size)

    def read_array(self, begin: int, end: int, size: Optional[int] = None) -> List[int]:
        """Read an array from the address."""
        if size is None:
            size = self.arch.addr_size

        data = self.read_bytes(begin, end - begin)
        array = []

        for offset in range(0, len(data), size):
            int_bytes = data[offset : offset + size]
            value = int.from_bytes(int_bytes, byteorder=self.endian)

            array.append(value)

        return array

    def write_int(self, address: int, value: int, size: int, signed: bool = False):
        """Write an integer into the address."""
        self.uc.mem_write(
            address,
            value.to_bytes(size, signed=signed, byteorder=self.endian),
        )

    def write_s8(self, address: int, value: int):
        """Write a signed int8 into the address."""
        self.write_int(address, value, 1, True)

    def write_s16(self, address: int, value: int):
        """Write a signed int16 into the address."""
        self.write_int(address, value, 2, True)

    def write_s32(self, address: int, value: int):
        """Write a signed int32 into the address."""
        self.write_int(address, value, 4, True)

    def write_s64(self, address: int, value: int):
        """Write a signed int64 into the address."""
        self.write_int(address, value, 8, True)

    def write_u8(self, address: int, value: int):
        """Write an unsigned int8 into the address."""
        self.write_int(address, value, 1, False)

    def write_u16(self, address: int, value: int):
        """Write an unsigned int16 into the address."""
        self.write_int(address, value, 2, False)

    def write_u32(self, address: int, value: int):
        """Write an unsigned int32 into the address."""
        self.write_int(address, value, 4, False)

    def write_u64(self, address: int, value: int):
        """Write an unsigned int64 into the address."""
        self.write_int(address, value, 8, False)

    def write_bytes(self, address: int, data: bytes):
        """Write bytes into the address."""
        self.uc.mem_write(address, data)

    def write_string(self, address: int, string: str):
        """Write string into the address."""
        self.uc.mem_write(address, string.encode("utf-8") + b"\x00")

    def write_pointer(self, address: int, value: int):
        """Write a pointer into the address."""
        self.write_int(address, value, self.arch.addr_size)

    def write_array(self, address: int, array: List[int], size: Optional[int] = None):
        """Write an array into the address."""
        if size is None:
            size = self.arch.addr_size

        data = b""

        for value in array:
            data += value.to_bytes(size, byteorder=self.endian)

        self.write_bytes(address, data)

    def call_symbol(self, symbol_name: str, *args: int) -> int:
        """Call function with the symbol name."""
        self.logger.info('Call symbol "%s"', symbol_name)

        symbol = self.find_symbol(symbol_name)
        address = symbol.address

        self._start_emulate(address, *args)

        return self.get_retval()

    def call_address(self, address: int, *args: int) -> int:
        """Call function at the address."""
        self._start_emulate(address, *args)

        return self.get_retval()



import pytest

from chomper.utils import aligned


@pytest.mark.parametrize(
    "x,n,expected",
    [
        (0, 1024, 0),
        (1, 1024, 1024),
        (1024, 1024, 1024),
        (1025, 1024, 2048),
    ],
)
def test_aligned(x, n, expected):
    result = aligned(x, n)

    assert result == expected



import pytest

from chomper.exceptions import SymbolMissingException


class TestCStandardLibrary:
    emu_names = ["emu_arm", "emu_arm64", "emu_ios"]

    @staticmethod
    def call_symbol(emu, emu_name, symbol_name, *args):
        if emu_name == "emu_ios":
            symbol_name = f"_{symbol_name}"

            try:
                emu.find_symbol(symbol_name)
            except SymbolMissingException:
                symbol_name = f"__platform{symbol_name}"

        return emu.call_symbol(symbol_name, *args)

    @pytest.mark.usefixtures("libc_arm", "libc_arm64")
    @pytest.mark.parametrize("emu_name", emu_names)
    def test_malloc(self, request, emu_name):
        emu = request.getfixturevalue(emu_name)

        result = self.call_symbol(emu, emu_name, "malloc", 16)

        assert result != 0

    @pytest.mark.usefixtures("libc_arm", "libc_arm64")
    @pytest.mark.parametrize("emu_name", emu_names)
    def test_free(self, request, emu_name):
        emu = request.getfixturevalue(emu_name)

        addr = self.call_symbol(emu, emu_name, "malloc", 16)
        self.call_symbol(emu, emu_name, "free", addr)

    @pytest.mark.usefixtures("libc_arm", "libc_arm64")
    @pytest.mark.parametrize("emu_name", emu_names)
    def test_memcpy(self, request, emu_name, str_test):
        emu = request.getfixturevalue(emu_name)

        v1 = emu.create_buffer(16)
        v2 = emu.create_string(str_test)

        self.call_symbol(emu, emu_name, "memcpy", v1, v2, len(str_test) + 1)
        result = emu.read_string(v1)

        assert result == str_test

    @pytest.mark.usefixtures("libc_arm", "libc_arm64")
    @pytest.mark.parametrize("emu_name", emu_names)
    def test_memcmp(self, request, emu_name, str_test):
        emu = request.getfixturevalue(emu_name)

        v1 = emu.create_string(str_test)
        v2 = emu.create_string(str_test)
        v3 = emu.create_string(str_test[::-1])

        result = self.call_symbol(emu, emu_name, "memcmp", v1, v2, len(str_test))

        assert result == 0

        result = self.call_symbol(emu, emu_name, "memcmp", v1, v3, len(str_test))

        assert result != 0

    @pytest.mark.usefixtures("libc_arm", "libc_arm64")
    @pytest.mark.parametrize("emu_name", emu_names)
    def test_memset(self, request, emu_name, str_test):
        emu = request.getfixturevalue(emu_name)

        size = len(str_test)

        v1 = emu.create_string(str_test)

        self.call_symbol(emu, emu_name, "memset", v1, 0, size)
        result = emu.read_bytes(v1, size)

        assert result == b"\x00" * size

    @pytest.mark.usefixtures("libc_arm", "libc_arm64")
    @pytest.mark.parametrize("emu_name", emu_names)
    def test_strncpy(self, request, emu_name, str_test):
        emu = request.getfixturevalue(emu_name)

        v1 = emu.create_buffer(16)
        v2 = emu.create_string(str_test)
        v3 = 5

        self.call_symbol(emu, emu_name, "strncpy", v1, v2, v3)
        result = emu.read_string(v1)

        assert result == str_test[:v3]

    @pytest.mark.usefixtures("libc_arm", "libc_arm64")
    @pytest.mark.parametrize("emu_name", emu_names)
    def test_strncmp(self, request, emu_name, str_test):
        emu = request.getfixturevalue(emu_name)

        v1 = emu.create_string(str_test)
        v2 = emu.create_string(str_test[:-1] + " ")

        result = self.call_symbol(emu, emu_name, "strncmp", v1, v2, len(str_test) - 1)

        assert result == 0

        result = self.call_symbol(emu, emu_name, "strncmp", v1, v2, len(str_test))

        assert result != 0

    @pytest.mark.usefixtures("libc_arm", "libc_arm64")
    @pytest.mark.parametrize("emu_name", emu_names)
    def test_strncat(self, request, emu_name, str_test):
        emu = request.getfixturevalue(emu_name)

        v1 = emu.create_buffer(32)
        v2 = emu.create_string(str_test)
        v3 = 5

        emu.write_string(v1, str_test)

        self.call_symbol(emu, emu_name, "strncat", v1, v2, v3)
        result = emu.read_string(v1)

        assert result == str_test + str_test[:v3]

    @pytest.mark.usefixtures("libc_arm", "libc_arm64")
    @pytest.mark.parametrize("emu_name", emu_names)
    def test_strcpy(self, request, emu_name, str_test):
        emu = request.getfixturevalue(emu_name)

        v1 = emu.create_buffer(16)
        v2 = emu.create_string(str_test)

        self.call_symbol(emu, emu_name, "strcpy", v1, v2)
        result = emu.read_string(v1)

        assert result == str_test

    @pytest.mark.usefixtures("libc_arm", "libc_arm64")
    @pytest.mark.parametrize("emu_name", emu_names)
    def test_strcmp(self, request, emu_name, str_test):
        emu = request.getfixturevalue(emu_name)

        v1 = emu.create_string(str_test)
        v2 = emu.create_string(str_test)
        v3 = emu.create_string(str_test[:-1] + " ")

        result = self.call_symbol(emu, emu_name, "strcmp", v1, v2)

        assert result == 0

        result = self.call_symbol(emu, emu_name, "strcmp", v1, v3)

        assert result != 0

    @pytest.mark.usefixtures("libc_arm", "libc_arm64")
    @pytest.mark.parametrize("emu_name", emu_names)
    def test_strcat(self, request, emu_name, str_test):
        emu = request.getfixturevalue(emu_name)

        v1 = emu.create_buffer(32)
        v2 = emu.create_string(str_test)

        emu.write_string(v1, str_test)

        self.call_symbol(emu, emu_name, "strcat", v1, v2)
        result = emu.read_string(v1)

        assert result == str_test + str_test

    @pytest.mark.usefixtures("libc_arm", "libc_arm64")
    @pytest.mark.parametrize("emu_name", emu_names)
    def test_strlen(self, request, emu_name, str_test):
        emu = request.getfixturevalue(emu_name)

        v1 = emu.create_string(str_test)

        result = self.call_symbol(emu, emu_name, "strlen", v1)

        assert result == len(str_test)

    @pytest.mark.usefixtures("libc_arm", "libc_arm64")
    @pytest.mark.parametrize("emu_name", ["emu_arm", "emu_arm64"])
    def test_sprintf(self, request, emu_name, str_test):
        emu = request.getfixturevalue(emu_name)

        fmt = "%d%s"

        v1 = emu.create_buffer(16)
        v2 = emu.create_string(fmt)
        v3 = len(str_test)
        v4 = emu.create_string(str_test)

        self.call_symbol(emu, emu_name, "sprintf", v1, v2, v3, v4)
        result = emu.read_string(v1)

        assert result == f"{len(str_test)}{str_test}"

    @pytest.mark.usefixtures("libc_arm", "libc_arm64")
    @pytest.mark.parametrize("emu_name", ["emu_arm", "emu_arm64"])
    def test_printf(self, request, emu_name, str_test):
        emu = request.getfixturevalue(emu_name)

        fmt = "%d%s"

        v1 = emu.create_string(fmt)
        v2 = len(str_test)
        v3 = emu.create_string(str_test)

        self.call_symbol(emu, emu_name, "printf", v1, v2, v3)



import os
import random

from chomper.types import FuncHooks


class CStandardLibraryHooks(FuncHooks):
    """Hooks for `libc.so`."""

    @classmethod
    def register(cls, emu):
        hooks = {
            "malloc": cls.hook_malloc,
            "free": cls.hook_free,
            "getpid": cls.hook_getpid,
            "arc4random": cls.hook_arc4random,
            "nanosleep": cls.hook_retval(0),
            "clock_nanosleep": cls.hook_retval(0),
            "pthread_mutex_lock": cls.hook_retval(0),
            "pthread_mutex_unlock": cls.hook_retval(0),
            "__ctype_get_mb_cur_max": cls.hook_retval(1),
        }

        emu.hooks.update(hooks)

    @staticmethod
    def hook_malloc(uc, address, size, user_data):
        emu = user_data["emu"]

        size = emu.get_arg(0)
        addr = emu.memory_manager.alloc(size)

        return addr

    @staticmethod
    def hook_free(uc, address, size, user_data):
        emu = user_data["emu"]

        addr = emu.get_arg(0)
        emu.memory_manager.free(addr)

    @staticmethod
    def hook_getpid(uc, address, size, user_data):
        return os.getpid()

    @staticmethod
    def hook_arc4random(uc, address, size, user_data):
        return random.randint(0, 0x100000000)



from chomper.types import Options


class IosOptions(Options):
    """Options of the iOS environment.

    Args:
        enable_objc: Enable Objective-C support. Defaults to True.
        enable_ui_kit: Enable UIKit support. Defaults to False.
    """

    def __init__(self, enable_objc: bool = True, enable_ui_kit: bool = False):
        self.enable_objc = enable_objc
        self.enable_ui_kit = enable_ui_kit



from typing import List

from unicorn import arm64_const, arm_const


class Arch:
    def __init__(
        self,
        nbits: int,
        reg_pc: int,
        reg_sp: int,
        reg_fp: int,
        reg_lr: int,
        reg_args: List[int],
        reg_retval: int,
    ):
        self.nbits = nbits

        self.reg_pc = reg_pc
        self.reg_sp = reg_sp
        self.reg_fp = reg_fp
        self.reg_lr = reg_lr

        self.reg_args = reg_args
        self.reg_retval = reg_retval

    @property
    def addr_size(self) -> int:
        return self.nbits // 8


arm_arch = Arch(
    nbits=32,
    reg_pc=arm_const.UC_ARM_REG_PC,
    reg_sp=arm_const.UC_ARM_REG_SP,
    reg_fp=arm_const.UC_ARM_REG_FP,
    reg_lr=arm_const.UC_ARM_REG_LR,
    reg_args=[
        arm_const.UC_ARM_REG_R0,
        arm_const.UC_ARM_REG_R1,
        arm_const.UC_ARM_REG_R2,
        arm_const.UC_ARM_REG_R3,
    ],
    reg_retval=arm_const.UC_ARM_REG_R0,
)

arm64_arch = Arch(
    nbits=64,
    reg_pc=arm64_const.UC_ARM64_REG_PC,
    reg_sp=arm64_const.UC_ARM64_REG_SP,
    reg_fp=arm64_const.UC_ARM64_REG_FP,
    reg_lr=arm64_const.UC_ARM64_REG_LR,
    reg_args=[
        arm64_const.UC_ARM64_REG_X0,
        arm64_const.UC_ARM64_REG_X1,
        arm64_const.UC_ARM64_REG_X2,
        arm64_const.UC_ARM64_REG_X3,
        arm64_const.UC_ARM64_REG_X4,
        arm64_const.UC_ARM64_REG_X5,
        arm64_const.UC_ARM64_REG_X6,
        arm64_const.UC_ARM64_REG_X7,
    ],
    reg_retval=arm64_const.UC_ARM64_REG_X0,
)



import os
from typing import Dict, List, Tuple, Optional

import lief
from lief.MachO import BIND_TYPES, ARM64_RELOCATION

from chomper.types import BaseLoader, Module, Symbol, Binding
from chomper.utils import aligned


class MachoLoader(BaseLoader):
    """The Mach-O file loader."""

    def _map_segments(self, binary: lief.MachO.Binary, module_base: int) -> int:
        """Map all segments into memory."""
        boundary = 0

        for segment in binary.segments:
            if not segment.virtual_size:
                continue

            seg_addr = module_base + segment.virtual_address

            map_addr = aligned(seg_addr, 1024) - (1024 if seg_addr % 1024 else 0)
            map_size = aligned(seg_addr - map_addr + segment.virtual_size, 1024)

            self.emu.uc.mem_map(map_addr, map_size)
            self.emu.uc.mem_write(
                seg_addr,
                bytes(segment.content),
            )

            boundary = max(boundary, map_addr + map_size)

        return boundary - module_base

    def _load_symbols(
        self,
        binary: lief.MachO.Binary,
        module_base: int,
    ) -> List[Symbol]:
        """Get all symbols in the module."""
        symbols = []

        lazy_bindings = self.get_lazy_bindings()
        lazy_binding_set = set()

        for symbol in binary.symbols:
            if symbol.value:
                symbol_name = str(symbol.name)
                symbol_address = module_base + symbol.value

                binding_name = symbol_name.replace("$VARIANT$armv81", "")

                # Lazy bind
                if lazy_bindings.get(binding_name):
                    # Avoid duplicate bind for special case like xx$VARIANT$armv81
                    if binding_name in lazy_binding_set and binding_name == symbol_name:
                        continue

                    for module, binding in lazy_bindings[binding_name]:
                        reloc_addr = symbol_address

                        if reloc_addr:
                            addr = module.base - module.image_base + binding.address

                            value = reloc_addr + binding.addend
                            value &= 0xFFFFFFFFFFFFFFFF

                            self.emu.write_pointer(addr, value)

                    lazy_binding_set.add(binding_name)

                symbol_struct = Symbol(
                    address=symbol_address,
                    name=symbol_name,
                )
                symbols.append(symbol_struct)

        return symbols

    def _get_symbol_reloc_addr(
        self,
        symbol_map: Dict[str, Symbol],
        symbol_name: str,
    ) -> Optional[int]:
        """Get the relocation address of a symbol."""
        reloc_addr = None

        if symbol_map.get(f"{symbol_name}$VARIANT$armv81"):
            if not self.emu.hooks.get(symbol_name):
                reloc_addr = symbol_map[f"{symbol_name}$VARIANT$armv81"].address

        elif symbol_name == "___chkstk_darwin":
            if symbol_map.get(f"_{symbol_name}"):
                reloc_addr = symbol_map[f"_{symbol_name}"].address

        elif symbol_map.get(symbol_name):
            reloc_addr = symbol_map[symbol_name].address

        elif symbol_map.get(f"__platform{symbol_name}"):
            reloc_addr = symbol_map[f"__platform{symbol_name}"].address

        return reloc_addr

    def _process_symbol_relocation(
        self,
        binary: lief.MachO.Binary,
        module_base: int,
        symbols: List[Symbol],
    ) -> List[Binding]:
        """Process relocations for symbols."""
        symbol_map = self.get_loaded_symbols()
        symbol_map.update({symbol.name: symbol for symbol in symbols})

        hooks_map: Dict = {}
        lazy_bindings = []

        for binding in binary.dyld_info.bindings:
            if binding.binding_type == BIND_TYPES.POINTER:
                symbol = binding.symbol
                symbol_name = str(symbol.name)

                reloc_addr = self._get_symbol_reloc_addr(symbol_map, symbol_name)

                if not reloc_addr:
                    lazy_binding = Binding(
                        symbol=symbol_name,
                        address=binding.address,
                        addend=binding.addend,
                    )
                    lazy_bindings.append(lazy_binding)

                    # Hook imports
                    if self.emu.hooks.get(symbol_name):
                        if not hooks_map.get(symbol_name):
                            reloc_addr = self.emu.create_buffer(self.emu.arch.addr_size)
                            hooks_map[symbol_name] = reloc_addr

                            self.emu.add_interceptor(
                                reloc_addr, self.emu.hooks[symbol_name]
                            )

                        else:
                            reloc_addr = hooks_map[symbol_name]

                        self.emu.logger.info(
                            'Hook import symbol "{}" at {}'.format(
                                symbol_name,
                                self.emu.debug_symbol(symbol.binding_info.address),
                            )
                        )

                    else:
                        # self.emu.add_hook(
                        #     reloc_addr,
                        #     self._missing_symbol_required_callback,
                        #     user_data={"symbol_name": symbol.name},
                        # )
                        continue

                if reloc_addr:
                    value = (reloc_addr + binding.addend) & 0xFFFFFFFFFFFFFFFF
                    self.emu.write_pointer(module_base + binding.address, value)

        return lazy_bindings

    def _process_relocation(
        self,
        binary: lief.MachO.Binary,
        module_base: int,
        symbols: List[Symbol],
    ):
        """Process relocations base on relocation table and symbol references."""
        blocks: List[Tuple[int, int]] = []

        begin = None
        end = None

        # Merge relocation records into blocks
        for segment in binary.segments:
            for relocation in segment.relocations:
                if relocation.type != ARM64_RELOCATION.SUBTRACTOR:
                    continue

                address = module_base + relocation.address

                if not begin:
                    begin = address

                if end and address != end:
                    blocks.append((begin, end))
                    begin = address

                end = address + self.emu.arch.addr_size

        # Read and write as blocks
        for begin, end in blocks:
            values = self.emu.read_array(begin, end)
            values = map(lambda v: module_base + v, values)

            self.emu.write_array(begin, values)

        return self._process_symbol_relocation(binary, module_base, symbols)

    def _get_init_array(self, binary: lief.MachO.Binary, module_base: int):
        """Get initialization functions in section `__mod_init_func`."""
        section = binary.get_section("__mod_init_func")
        if not section:
            return []

        begin = module_base + section.virtual_address
        end = begin + section.size

        values = self.emu.read_array(begin, end)

        return [addr for addr in values if addr]

    def load(
        self,
        module_base: int,
        module_file: str,
        trace_symbol_calls: bool = False,
    ) -> Module:
        """Load Mach-O executable file from path."""
        module_name = os.path.basename(module_file)
        self.emu.logger.info(f'Load module "{module_name}"')

        binary: lief.MachO.Binary = lief.parse(module_file)  # type: ignore

        segment_addresses = [t.virtual_address for t in binary.segments]

        # Make module memory distribution more compact
        image_base = aligned(min(segment_addresses), 1024)
        module_base -= image_base

        size = self._map_segments(binary, module_base)
        symbols = self._load_symbols(binary, module_base)

        self.add_symbol_hooks(symbols, trace_symbol_calls)

        lazy_bindings = self._process_relocation(binary, module_base, symbols)
        init_array = self._get_init_array(binary, module_base)

        return Module(
            base=module_base + image_base,
            size=size - image_base,
            name=module_name,
            symbols=symbols,
            init_array=init_array,
            image_base=image_base,
            lazy_bindings=lazy_bindings,
            binary=binary,
        )



import datetime
import random
import time
import uuid

from chomper.types import FuncHooks


class IosFuncHooks(FuncHooks):
    """System function hooks for the iOS."""

    @classmethod
    def register(cls, emu):
        retval_hooks = {
            # libsystem_pthread.dylib
            "_pthread_self": 1,
            "_pthread_rwlock_rdlock": 0,
            "_pthread_rwlock_unlock": 0,
            # libsystem_platform.dylib
            "_os_unfair_lock_assert_owner": None,
            # libsystem_info.dylib
            "_getpwuid": 0,
            "_getpwuid_r": 0,
            # libsystem_trace.dylib
            "__os_activity_initiate": 0,
            # libsystem_notify.dylib
            "_notify_register_dispatch": 0,
            # libdispatch.dylib
            "_dispatch_async": 0,
            # libdyld.dylib
            "_dyld_program_sdk_at_least": 1,
            # libobjc.A.dylib
            "__ZN11objc_object16rootAutorelease2Ev": None,
            # CoreFoundation
            "__CFBundleCreateInfoDictFromMainExecutable": 0,
            # Foundation
            "_NSLog": 0,
        }

        hooks = {
            # libsystem_malloc.dylib
            "_malloc": cls.hook_malloc,
            "_calloc": cls.hook_calloc,
            "_realloc": cls.hook_realloc,
            "_free": cls.hook_free,
            "_malloc_size": cls.hook_malloc_size,
            "_malloc_default_zone": cls.hook_malloc_default_zone,
            "_malloc_zone_malloc": cls.hook_malloc_zone_malloc,
            "_malloc_zone_calloc": cls.hook_malloc_zone_calloc,
            "_malloc_zone_realloc": cls.hook_malloc_zone_realloc,
            "_malloc_zone_free": cls.hook_malloc_zone_free,
            "_malloc_zone_from_ptr": cls.hook_malloc_zone_from_ptr,
            "_malloc_zone_memalign": cls.hook_malloc_zone_memalign,
            "_malloc_good_size": cls.hook_malloc_good_size,
            "_malloc_engaged_nano": cls.hook_malloc_engaged_nano,
            "_posix_memalign": cls.hook_posix_memalign,
            # libsystem_c.dylib
            "_getcwd": cls.hook_getcwd,
            "_opendir": cls.hook_opendir,
            "_time": cls.hook_time,
            "_srandom": cls.hook_srandom,
            "_random": cls.hook_random,
            "_localtime_r": cls.hook_localtime_r,
            "___srefill": cls.hook_srefill,
            # libsystem_kernel.dylib
            "_stat": cls.hook_stat,
            "_lstat": cls.hook_lstat,
            "___sysctlbyname": cls.hook_sysctlbyname,
            # libmacho.dylib
            "_getsectiondata": cls.hook_getsectiondata,
            "_getsegmentdata": cls.hook_getsegmentdata,
        }

        for func_name, retval in retval_hooks.items():
            hooks[func_name] = cls.hook_retval(retval)

        emu.hooks.update(hooks)

    @staticmethod
    def hook_malloc(uc, address, size, user_data):
        emu = user_data["emu"]

        size = emu.get_arg(0)
        mem = emu.memory_manager.alloc(size)

        return mem

    @staticmethod
    def hook_calloc(uc, address, size, user_data):
        emu = user_data["emu"]

        numitems = emu.get_arg(0)
        size = emu.get_arg(1)

        mem = emu.memory_manager.alloc(numitems * size)
        emu.write_bytes(mem, b"\x00" * (numitems * size))

        return mem

    @staticmethod
    def hook_realloc(uc, address, size, user_data):
        emu = user_data["emu"]

        ptr = emu.get_arg(0)
        size = emu.get_arg(1)

        return emu.memory_manager.realloc(ptr, size)

    @staticmethod
    def hook_free(uc, address, size, user_data):
        emu = user_data["emu"]

        mem = emu.get_arg(0)
        emu.memory_manager.free(mem)

    @staticmethod
    def hook_malloc_size(uc, address, size, user_data):
        emu = user_data["emu"]

        mem = emu.get_arg(0)

        for pool in emu.memory_manager.pools:
            if pool.address <= mem < pool.address + pool.size:
                return pool.block_size

        return 0

    @staticmethod
    def hook_malloc_default_zone(uc, address, size, user_data):
        return 0

    @staticmethod
    def hook_malloc_zone_malloc(uc, address, size, user_data):
        emu = user_data["emu"]

        size = emu.get_arg(1)
        mem = emu.memory_manager.alloc(size)

        return mem

    @staticmethod
    def hook_malloc_zone_calloc(uc, address, size, user_data):
        emu = user_data["emu"]

        numitems = emu.get_arg(1)
        size = emu.get_arg(2)

        mem = emu.memory_manager.alloc(numitems * size)
        emu.write_bytes(mem, b"\x00" * (numitems * size))

        return mem

    @staticmethod
    def hook_malloc_zone_realloc(uc, address, size, user_data):
        emu = user_data["emu"]

        ptr = emu.get_arg(1)
        size = emu.get_arg(2)

        return emu.memory_manager.realloc(ptr, size)

    @staticmethod
    def hook_malloc_zone_free(uc, address, size, user_data):
        emu = user_data["emu"]

        mem = emu.get_arg(1)
        emu.memory_manager.free(mem)

    @staticmethod
    def hook_malloc_zone_from_ptr(uc, address, size, user_data):
        return 0

    @staticmethod
    def hook_malloc_zone_memalign(uc, address, size, user_data):
        emu = user_data["emu"]

        size = emu.get_arg(2)
        mem = emu.memory_manager.alloc(size)

        return mem

    @staticmethod
    def hook_malloc_good_size(uc, address, size, user_data):
        emu = user_data["emu"]

        size = emu.get_arg(0)

        return size

    @staticmethod
    def hook_malloc_engaged_nano(uc, address, size, user_data):
        return 1

    @staticmethod
    def hook_posix_memalign(uc, address, size, user_data):
        emu = user_data["emu"]

        memptr = emu.get_arg(0)
        size = emu.get_arg(2)

        mem = emu.memory_manager.alloc(size)
        emu.write_pointer(memptr, mem)

        return 0

    @staticmethod
    def hook_os_unfair_lock_assert_owner(uc, address, size, user_data):
        pass

    @staticmethod
    def hook_srefill(uc, address, size, user_data):
        return 0

    @staticmethod
    def hook_getcwd(uc, address, size, user_data):
        emu = user_data["emu"]

        path = f"/private/var/containers/Bundle/Application/{uuid.uuid4()}/"

        buf = emu.get_arg(0)
        emu.write_string(buf, path)

        return 0

    @staticmethod
    def hook_opendir(uc, address, size, user_data):
        return 0

    @staticmethod
    def hook_time(uc, address, size, user_data):
        return int(time.time())

    @staticmethod
    def hook_srandom(uc, address, size, user_data):
        return 0

    @staticmethod
    def hook_random(uc, address, size, user_data):
        return random.randint(0, 2**32 - 1)

    @staticmethod
    def hook_localtime_r(uc, address, size, user_data):
        emu = user_data["emu"]

        tp = emu.read_u64(emu.get_arg(0))
        tm = emu.get_arg(1)

        date = datetime.datetime.fromtimestamp(tp)

        emu.write_u32(tm, date.second)
        emu.write_u32(tm + 4, date.minute)
        emu.write_u32(tm + 8, date.hour)
        emu.write_u32(tm + 12, date.day)
        emu.write_u32(tm + 16, date.month)
        emu.write_u32(tm + 20, date.year)
        emu.write_u32(tm + 24, 0)
        emu.write_u32(tm + 28, 0)
        emu.write_u32(tm + 32, 0)
        emu.write_u64(tm + 40, 8 * 60 * 60)

        return 0

    @staticmethod
    def hook_fopen(uc, address, size, user_data):
        emu = user_data["emu"]

        path = emu.read_string(emu.get_arg(0))
        emu.logger.info("fopen: %s" % path)

        return 0

    @staticmethod
    def hook_stat(uc, address, size, user_data):
        emu = user_data["emu"]

        stat = emu.get_arg(1)

        # st_mode
        emu.write_u32(stat + 4, 0x4000)

        return 0

    @staticmethod
    def hook_lstat(uc, address, size, user_data):
        emu = user_data["emu"]

        stat = emu.get_arg(1)

        # st_mode
        emu.write_u32(stat + 4, 0x4000)

        return 0

    @staticmethod
    def hook_sysctlbyname(uc, address, size, user_data):
        emu = user_data["emu"]

        name = emu.read_string(emu.get_arg(0))
        oldp = emu.get_arg(2)
        oldlenp = emu.get_arg(3)

        if not oldp or not oldlenp:
            return 0

        if name == "kern.boottime":
            emu.write_u64(oldp, int(time.time()) - 3600 * 24)
            emu.write_u64(oldp + 8, 0)

        elif name == "kern.osvariant_status":
            # internal_release_type = 3
            emu.write_u64(oldp, 0x30)

        elif name == "hw.memsize":
            emu.write_u64(oldp, 4 * 1024 * 1024 * 1024)

        else:
            emu.logger.info([emu.debug_symbol(t[0]) for t in emu.backtrace()])
            raise RuntimeError("Unhandled sysctl command: %s" % name)

        return 0

    @staticmethod
    def hook_getsectiondata(uc, address, size, user_data):
        emu = user_data["emu"]
        module = emu.modules[-1]

        section_name = emu.read_string(emu.get_arg(2))
        size_ptr = emu.get_arg(3)

        section = module.binary.get_section(section_name)
        if not section:
            return 0

        emu.write_u64(size_ptr, section.size)

        return module.base - module.image_base + section.virtual_address

    @staticmethod
    def hook_getsegmentdata(uc, address, size, user_data):
        emu = user_data["emu"]
        module = emu.modules[-1]

        segment_name = emu.read_string(emu.get_arg(1))
        size_ptr = emu.get_arg(2)

        segment = module.binary.get_segment(segment_name)
        if not segment:
            return 0

        emu.write_u64(size_ptr, segment.virtual_size)

        return module.base - module.image_base + segment.virtual_address



import os
from ctypes import sizeof
from typing import List

from . import modules
from .hooks import IosFuncHooks
from .loader import MachoLoader
from .options import IosOptions
from .structs import MachHeader64
from .syscall import IosSyscallHandler

from chomper.types import BaseOs, Module
from chomper.utils import struct2bytes


class IosOs(BaseOs):
    """The iOS environment."""

    def __init__(self, emu, **kwargs):
        super().__init__(emu, **kwargs)

        if not self.options:
            self.options = IosOptions()

        self.loader = MachoLoader(emu)

    def init_hooks(self):
        """Initialize the hooks."""
        IosFuncHooks.register(self.emu)

    def init_syscall_handlers(self):
        """Initialize the system call handlers."""
        IosSyscallHandler.register(self.emu)

    def find_system_module(self, module_name: str) -> str:
        """Find system module in rootfs directory.

        raises:
            FileNotFoundError: If module not found.
        """
        lib_dirs = [
            "usr/lib/system",
            "usr/lib",
            "System/Library/Frameworks",
            "System/Library/PrivateFrameworks",
        ]

        for lib_dir in lib_dirs:
            path = os.path.join(self.rootfs_path or ".", lib_dir)

            lib_path = os.path.join(path, module_name)
            if os.path.exists(lib_path):
                return lib_path

            framework_path = os.path.join(path, f"{module_name}.framework")
            if os.path.exists(framework_path):
                return os.path.join(framework_path, module_name)

        raise FileNotFoundError("Module '%s' not found" % module_name)

    def check_module(self, module_name: str) -> bool:
        """Check if the module exist."""
        return module_name in (t.name for t in self.emu.modules)

    def resolve_modules(self, module_classes: List):
        """Load modules if don't exist."""
        for module_class in module_classes:
            if self.check_module(module_class.MODULE_NAME):
                continue

            module = module_class(self.emu)
            module.load()

    def init_objc(self, module: Module):
        """Initialize Objective-C."""
        if not self.check_module(modules.Objc.MODULE_NAME):
            return

        if not module.binary:
            return

        try:
            mach_header = MachHeader64(
                magic=module.binary.header.magic.value,
                cputype=module.binary.header.cpu_type.value,
                cpusubtype=module.binary.header.cpu_subtype,
                filetype=module.binary.header.file_type.value,
                ncmds=module.binary.header.nb_cmds,
                sizeofcmds=module.binary.header.sizeof_cmds,
                flags=module.binary.header.flags,
                reserved=module.binary.header.reserved,
            )

            mach_header_ptr = self.emu.create_buffer(sizeof(MachHeader64))
            self.emu.write_bytes(mach_header_ptr, struct2bytes(mach_header))

            mach_header_ptrs = self.emu.create_buffer(self.emu.arch.addr_size)
            self.emu.write_pointer(mach_header_ptrs, mach_header_ptr)

            self.emu.call_symbol("_map_images", 1, 0, mach_header_ptrs)
            self.emu.call_symbol("_load_images")

        except Exception as e:
            self.emu.logger.error("Initialize Objective-C failed.")
            self.emu.logger.exception(e)

    def enable_objc(self):
        """Enable Objective-C support."""
        dependent_modules = [
            modules.SystemPlatform,
            modules.SystemKernel,
            modules.SystemC,
            modules.SystemPthread,
            modules.SystemInfo,
            modules.SystemDarwin,
            modules.SystemFeatureflags,
            modules.Corecrypto,
            modules.CommonCrypto,
            modules.Cppabi,
            modules.Cpp,
            modules.Dyld,
            modules.Objc,
            modules.Dispatch,
            modules.SystemBlocks,
            modules.SystemTrace,
            modules.SystemSandbox,
            modules.Network,
            modules.CoreFoundation,
            modules.CFNetwork,
            modules.Foundation,
            modules.Security,
        ]

        self.resolve_modules(dependent_modules)

    def enable_ui_kit(self):
        """Enable UIKit support.

        Mainly used to load `UIDevice` class, which is used to get device info.
        """
        dependent_modules = [
            modules.QuartzCore,
            modules.BaseBoard,
            modules.FrontBoardServices,
            modules.PrototypeTools,
            modules.TextInput,
            modules.PhysicsKit,
            modules.CoreAutoLayout,
            modules.UIFoundation,
            modules.UIKitServices,
            modules.UIKitCore,
        ]

        self.resolve_modules(dependent_modules)

    def initialize(self):
        """Initialize the environment."""
        self.init_hooks()
        self.init_syscall_handlers()

        if self.options.enable_objc:
            self.enable_objc()

        if self.options.enable_ui_kit:
            self.enable_ui_kit()



import pytest


class TestEmulator:
    @pytest.mark.usefixtures("libc_arm64")
    def test_find_symbol(self, emu_arm64):
        symbol_name = "malloc"
        symbol = emu_arm64.find_symbol(symbol_name)

        assert symbol.name == symbol_name

    def test_backtrace(self, emu_arm64, libtiny_v73021_arm64):
        def hook_code(uc, address, size, user_data):
            trace_stack = emu_arm64.backtrace()
            assert len(trace_stack) != 0

        emu_arm64.add_hook(libtiny_v73021_arm64.base + 0x2BA08, hook_code)

        a1 = emu_arm64.create_buffer(32)
        a2 = emu_arm64.create_buffer(32)
        a3 = emu_arm64.create_buffer(32)

        emu_arm64.call_address(libtiny_v73021_arm64.base + 0x289A4, a1, a2, a3)

    @pytest.mark.usefixtures("libz_arm64")
    def test_set_and_get_arg(self, emu_arm64):
        args = [i for i in range(16)]

        for index, arg in enumerate(args):
            emu_arm64.set_arg(index, arg)

        for index, arg in enumerate(args):
            assert emu_arm64.get_arg(index) == arg

    @pytest.mark.usefixtures("libz_arm64")
    def test_set_and_get_retval(self, emu_arm64):
        retval = 105

        emu_arm64.set_retval(retval)
        result = emu_arm64.get_retval()

        assert result == retval

    def test_create_buffer(self, emu_arm64):
        result = emu_arm64.create_buffer(1024)

        assert result is not None

    def test_create_string(self, emu_arm64, str_test):
        result = emu_arm64.create_string(str_test)

        assert result is not None

    def test_free(self, emu_arm64):
        addr = emu_arm64.create_buffer(1024)
        emu_arm64.free(addr)

    def test_read_and_write_int(self, emu_arm64):
        addr = emu_arm64.create_buffer(1024)
        value = 105

        emu_arm64.write_int(addr, value, size=4)
        result = emu_arm64.read_int(addr, size=4)

        assert result == value

    def test_read_and_write_bytes(self, emu_arm64, bytes_test):
        addr = emu_arm64.create_buffer(1024)

        emu_arm64.write_bytes(addr, bytes_test)
        result = emu_arm64.read_bytes(addr, len(bytes_test))

        assert result == bytes_test

    def test_read_and_write_string(self, emu_arm64, str_test):
        addr = emu_arm64.create_buffer(1024)

        emu_arm64.write_string(addr, str_test)
        result = emu_arm64.read_string(addr)

        assert result == str_test

    @pytest.mark.usefixtures("libz_arm64")
    def test_call_symbol(self, emu_arm64):
        result = emu_arm64.call_symbol("zlibVersion")

        assert emu_arm64.read_string(result) == "1.2.8"

    @pytest.mark.usefixtures("libz_arm64")
    def test_call_address(self, emu_arm64):
        symbol = emu_arm64.find_symbol("zlibVersion")
        result = emu_arm64.call_address(symbol.address)

        assert emu_arm64.read_string(result) == "1.2.8"

    def test_exec_init_array(self, emu_arm64, libszstone_v4945_arm64):
        assert emu_arm64.read_string(libszstone_v4945_arm64.base + 0x49DD8) == "1.2.3"



import zlib

import pytest


class TestExamples:
    @pytest.mark.usefixtures("libc_arm", "libz_arm")
    def test_libdusanwa_v4856_arm(self, emu_arm, libdusanwa_v4856_arm, bytes_test):
        a1 = emu_arm.create_buffer(32)
        a2 = 32
        a3 = emu_arm.create_buffer(32)
        a4 = emu_arm.create_buffer(32)

        emu_arm.write_bytes(a1, bytes_test)
        emu_arm.write_bytes(a4, bytes_test)

        emu_arm.call_address((libdusanwa_v4856_arm.base + 0xA588) | 1, a1, a2, a3, a4)
        result = emu_arm.read_bytes(a3, a2)

        assert zlib.crc32(result) == 1148403178

    @pytest.mark.usefixtures("libc_arm64", "libz_arm64")
    def test_libszstone_v4945_arm64(
        self, emu_arm64, libszstone_v4945_arm64, bytes_test
    ):
        a1 = emu_arm64.create_buffer(len(bytes_test))
        a2 = len(bytes_test)
        a3 = emu_arm64.create_buffer(1024)

        emu_arm64.write_bytes(a1, bytes_test)

        result_size = emu_arm64.call_address(
            libszstone_v4945_arm64.base + 0x2F1C8, a1, a2, a3
        )
        result = emu_arm64.read_bytes(a3, result_size)

        assert zlib.crc32(result) == 3884391316

    @pytest.mark.usefixtures("libc_arm64", "libz_arm64")
    def test_libtiny_v73021_arm64(self, emu_arm64, libtiny_v73021_arm64, bytes_test):
        a1 = emu_arm64.create_buffer(32)
        a2 = emu_arm64.create_buffer(32)
        a3 = emu_arm64.create_buffer(32)

        emu_arm64.write_bytes(a1, bytes_test * 4)
        emu_arm64.write_bytes(a2, bytes_test * 4)

        emu_arm64.call_address(libtiny_v73021_arm64.base + 0x289A4, a1, a2, a3)
        result = emu_arm64.read_bytes(a3, 32)

        assert zlib.crc32(result) == 4192995551

    # @staticmethod
    # def test_duapp_v581(emu_ios, sample_str):
    #     duapp = emu_ios.load_module("DUApp")
    #
    #     a1 = emu_ios.create_string("objc")
    #     a2 = emu_ios.create_string(sample_str)
    #     a3 = len(sample_str)
    #     a4 = emu_ios.create_string(str(uuid.uuid4()))
    #     a5 = emu_ios.create_buffer(8)
    #     a6 = emu_ios.create_buffer(8)
    #     a7 = emu_ios.create_string("com.siwuai.duapp")
    #
    #     emu_ios.call_address(duapp.base + 0x109322118, a1, a2, a3, a4, a5, a6, a7)
    #     result = emu_ios.read_string(emu_ios.read_pointer(a5))
    #
    #     assert re.match(r"\w{32}\.[\w=]+\.", result)



from dataclasses import dataclass
from enum import Enum
from typing import Dict, List, Optional, TypeVar

import lief


class SymbolType(Enum):
    FUNC = 1
    OTHER = 2


@dataclass
class Symbol:
    address: int
    name: str
    type: Optional[SymbolType] = None


@dataclass
class Binding:
    symbol: str
    address: int
    addend: int


@dataclass
class Module:
    base: int
    size: int
    name: str

    symbols: List[Symbol]

    init_array: Optional[List[int]] = None

    image_base: Optional[int] = None

    lazy_bindings: Optional[List[Binding]] = None

    binary: Optional[lief.MachO.Binary] = None


class FuncHooks:
    @classmethod
    def register(cls, emu):
        raise NotImplementedError

    @classmethod
    def hook_retval(cls, retval):
        def decorator(uc, address, size, user_data):
            return retval

        return decorator


class SyscallHandler:
    @classmethod
    def register(cls, emu):
        raise NotImplementedError


class BaseLoader:
    """Executable file loader."""

    def __init__(self, emu):
        self.emu = emu

    def get_loaded_symbols(self) -> Dict[str, Symbol]:
        """Get loaded symbols."""
        symbol_map = {}

        for module in self.emu.modules:
            for symbol in module.symbols:
                symbol_map[symbol.name] = symbol

        return symbol_map

    def get_lazy_bindings(self) -> Dict[str, List]:
        """Get lazy bindings."""
        bindings: Dict[str, List] = {}

        for module in self.emu.modules:
            for binding in module.lazy_bindings:
                symbol_name = binding.symbol

                if not bindings.get(symbol_name):
                    bindings[symbol_name] = []

                bindings[symbol_name].append((module, binding))

        return bindings

    def add_symbol_hooks(self, symbols: List[Symbol], trace_symbol_calls: bool = False):
        """Add hooks to target symbols."""
        for symbol in symbols:
            if symbol.type == SymbolType.OTHER:
                continue

            # Trace symbol call
            if trace_symbol_calls:
                self.emu.add_hook(
                    symbol.address,
                    self.emu.trace_symbol_call_callback,
                    user_data={"symbol": symbol},
                )

            # Hook symbol
            if self.emu.hooks.get(symbol.name):
                message = 'Hook export symbol "{}" at {}'.format(
                    symbol.name,
                    self.emu.debug_symbol(symbol.address),
                )
                self.emu.logger.info(message)

                self.emu.add_interceptor(symbol.address, self.emu.hooks[symbol.name])

    def load(
        self,
        module_base: int,
        module_file: str,
        trace_symbol_calls: bool = False,
    ) -> Module:
        """Load executable file."""
        raise NotImplementedError


class Options:
    pass


OptionsTV = TypeVar("OptionsTV", bound=Options)


class BaseOs:
    def __init__(
        self,
        emu,
        rootfs_path: Optional[str] = None,
        options: Optional[OptionsTV] = None,
    ):
        self.emu = emu
        self.rootfs_path = rootfs_path
        self.options = options

    def initialize(self):
        pass



from ctypes import Structure, c_uint32


class MachHeader64(Structure):
    _fields_ = [
        ("magic", c_uint32),
        ("cputype", c_uint32),
        ("cpusubtype", c_uint32),
        ("filetype", c_uint32),
        ("ncmds", c_uint32),
        ("sizeofcmds", c_uint32),
        ("flags", c_uint32),
        ("reserved", c_uint32),
    ]



from ctypes import addressof, create_string_buffer, sizeof, memmove, Structure


def aligned(x: int, n: int) -> int:
    """Align `x` based on `n`."""
    return ((x - 1) // n + 1) * n


def struct2bytes(st: Structure) -> bytes:
    """Convert struct to bytes."""
    buffer = create_string_buffer(sizeof(st))
    memmove(buffer, addressof(st), sizeof(st))
    return buffer.raw



import os
from typing import Dict, List

from elftools.elf.dynamic import DynamicSegment
from elftools.elf.elffile import ELFFile
from elftools.elf.enums import ENUM_RELOC_TYPE_AARCH64, ENUM_RELOC_TYPE_ARM
from elftools.elf.relocation import Relocation

from chomper.types import BaseLoader, Module, Symbol, SymbolType
from chomper.utils import aligned


class ELFLoader(BaseLoader):
    """The ELF file loader."""

    def _map_segments(self, elffile: ELFFile, module_base: int) -> int:
        """Map segment of type `LOAD` into memory."""
        boundary = 0

        for segment in elffile.iter_segments(type="PT_LOAD"):
            seg_addr = module_base + segment.header.p_vaddr

            # Because of alignment, the actual mapped address range will be larger.
            map_addr = aligned(seg_addr, 1024) - (1024 if seg_addr % 1024 else 0)
            map_size = aligned(seg_addr - map_addr + segment.header.p_memsz, 1024)

            self.emu.uc.mem_map(map_addr, map_size)
            self.emu.uc.mem_write(seg_addr, segment.data())

            boundary = max(boundary, map_addr + map_size)

        return boundary - module_base

    @staticmethod
    def _load_symbols(elffile: ELFFile, module_base: int) -> List[Symbol]:
        """Get all symbols in the module."""
        symbols = []

        for segment in elffile.iter_segments(type="PT_DYNAMIC"):
            for symbol in segment.iter_symbols():
                if symbol.entry["st_info"]["bind"] in ("STB_GLOBAL", "STB_WEAK"):
                    if symbol.entry["st_shndx"] == "SHN_UNDEF":
                        continue

                    symbol_type = (
                        SymbolType.FUNC
                        if symbol.entry["st_info"]["type"] == "STT_FUNC"
                        else SymbolType.OTHER
                    )

                    symbol_struct = Symbol(
                        address=module_base + symbol.entry["st_value"],
                        name=symbol.name,
                        type=symbol_type,
                    )
                    symbols.append(symbol_struct)

        return symbols

    def _relocate_address(
        self,
        module_base: int,
        symbol_map: Dict[str, Symbol],
        relocation: Relocation,
        segment: DynamicSegment,
    ):
        """Relocate address in the relocation table."""
        reloc_type = relocation["r_info_type"]
        reloc_offset = module_base + relocation["r_offset"]
        reloc_addr = None

        if reloc_type in (
            ENUM_RELOC_TYPE_ARM["R_ARM_ABS32"],
            ENUM_RELOC_TYPE_ARM["R_ARM_GLOB_DAT"],
            ENUM_RELOC_TYPE_ARM["R_ARM_JUMP_SLOT"],
            ENUM_RELOC_TYPE_AARCH64["R_AARCH64_ABS64"],
            ENUM_RELOC_TYPE_AARCH64["R_AARCH64_GLOB_DAT"],
            ENUM_RELOC_TYPE_AARCH64["R_AARCH64_JUMP_SLOT"],
        ):
            symbol_name = segment.get_symbol(relocation["r_info_sym"]).name

            if symbol_map.get(symbol_name):
                reloc_addr = symbol_map[symbol_name].address

            else:
                # If the symbol is missing, let it jump to a specified address and
                # the address has hooked to raise an exception with useful information.
                hook_addr = self.emu.create_buffer(self.emu.arch.addr_size)

                self.emu.add_hook(
                    hook_addr,
                    self.emu.missing_symbol_required_callback,
                    user_data={"symbol_name": symbol_name},
                )
                self.emu.write_pointer(reloc_offset, hook_addr)

        elif reloc_type in (
            ENUM_RELOC_TYPE_ARM["R_ARM_RELATIVE"],
            ENUM_RELOC_TYPE_AARCH64["R_AARCH64_RELATIVE"],
        ):
            if relocation.is_RELA():
                reloc_addr = module_base + relocation["r_addend"]

        if reloc_addr:
            self.emu.write_pointer(reloc_offset, reloc_addr)

    def _process_relocation(
        self,
        elffile: ELFFile,
        module_base: int,
        symbols: List[Symbol],
    ):
        """Process relocation tables."""
        symbol_map = self.get_loaded_symbols()
        symbol_map.update({symbol.name: symbol for symbol in symbols})

        for segment in elffile.iter_segments(type="PT_DYNAMIC"):
            for relocation_table in segment.get_relocation_tables().values():
                for relocation in relocation_table.iter_relocations():
                    self._relocate_address(module_base, symbol_map, relocation, segment)

    def _get_init_array(self, elffile: ELFFile, module_base: int) -> List[int]:
        """Get initialization functions in section `.init_array`."""
        init_funcs = []

        init_array_addr = None
        init_array_size = None

        for segment in elffile.iter_segments(type="PT_DYNAMIC"):
            for tag in segment.iter_tags():
                if tag.entry.d_tag == "DT_INIT_ARRAY":
                    init_array_addr = tag.entry.d_val

                elif tag.entry.d_tag == "DT_INIT_ARRAYSZ":
                    init_array_size = tag.entry.d_val

        if init_array_addr is not None and init_array_size is not None:
            begin = module_base + init_array_addr
            end = begin + init_array_size

            init_funcs = self.emu.read_array(begin, end)

        return init_funcs

    def load(
        self,
        module_base: int,
        module_file: str,
        trace_symbol_calls: bool = False,
    ) -> Module:
        """Load ELF library file from path."""
        with ELFFile.load_from_path(module_file) as elffile:
            module_name = os.path.basename(module_file)
            self.emu.logger.info(f'Load module "{module_name}"')

            # Map segments into memory.
            size = self._map_segments(elffile, module_base)
            symbols = self._load_symbols(elffile, module_base)

            self.add_symbol_hooks(symbols, trace_symbol_calls)

            # Process address relocation.
            self._process_relocation(elffile, module_base, symbols)

            init_array = self._get_init_array(elffile, module_base)

            return Module(
                base=module_base,
                size=size,
                name=module_name,
                symbols=symbols,
                init_array=init_array,
            )



import logging
from typing import Optional

default_handler = logging.StreamHandler()
default_handler.setFormatter(
    logging.Formatter("%(asctime)s - %(name)s - %(levelname)s: %(message)s")
)


def get_logger(name: Optional[str] = None, level: int = logging.INFO) -> logging.Logger:
    """Get default logger."""
    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.addHandler(default_handler)
    return logger



from typing import List

from unicorn import Uc

from .utils import aligned


class MemoryPool:
    def __init__(self, address: int, size: int, blocks: List[int]):
        self.address = address
        self.size = size
        self.blocks = blocks

    @property
    def block_num(self) -> int:
        return len(self.blocks)

    @property
    def block_size(self) -> int:
        return self.size // self.block_num


class MemoryManager:
    """Provide memory management on Unicorn.

    Each mapped memory is treated as a pool, and each pool contains one or more
    blocks of equal size. When allocating, it will first iterate pools to find
    one with enough block size, and then return an unused block. If the requested
    size is greater than `minimum_pool_size`, it will create a poll with only
    one block and enough size.

    Args:
        uc: The Unicorn instance.
        address: Base address of mapped memory.
        minimum_pool_size: Minimum size of pool.
    """

    def __init__(self, uc: Uc, address: int, minimum_pool_size: int):
        self.uc = uc
        self.address = address
        self.minimum_pool_size = minimum_pool_size

        self.pools: List[MemoryPool] = []

        self.init_pools()

    def create_pool(self, block_size: int) -> MemoryPool:
        """Create a pool."""
        if block_size <= self.minimum_pool_size:
            pool = MemoryPool(
                address=self.address,
                size=self.minimum_pool_size,
                blocks=[0 for _ in range(self.minimum_pool_size // block_size)],
            )

        else:
            pool = MemoryPool(address=self.address, size=block_size, blocks=[0])

        self.uc.mem_map(pool.address, pool.size)
        self.address += pool.size

        return pool

    def init_pools(self):
        """Initialize pools.

        At the begging, some pools with different block sizes will be created.
        """
        block_sizes = [8 * 2**i for i in range(10)]

        for block_size in block_sizes:
            self.pools.append(self.create_pool(block_size))

    def alloc(self, size: int) -> int:
        """Allocate memory."""
        if size > self.minimum_pool_size:
            block_size = aligned(size, 1024)

        else:
            block_size = 8
            while block_size < size:
                block_size *= 2

        for pool in self.pools:
            if pool.block_size == block_size:
                try:
                    index = pool.blocks.index(0)
                    pool.blocks[index] = 1
                    return pool.address + index * pool.block_size

                except ValueError:
                    continue

        pool = self.create_pool(block_size)
        pool.blocks[0] = 1

        return pool.address

    def realloc(self, address: int, size: int) -> int:
        """Reallocate memory."""
        block_size = 0

        for pool in self.pools:
            if pool.address <= address < pool.address + pool.size:
                block_size = pool.block_size

        if block_size >= size:
            return address

        new_address = self.alloc(size)
        self.uc.mem_write(new_address, bytes(self.uc.mem_read(address, block_size)))

        return new_address

    def free(self, address: int):
        """Free memory."""
        for pool in self.pools:
            if pool.address <= address < pool.address + pool.size:
                index = (address - pool.address) // pool.block_size
                pool.blocks[index] = 0



import os
import random
import time
import threading

from chomper.types import SyscallHandler


class IosSyscallNo:
    """System call numbers in iOS."""

    GETPID = 0x14
    GETUID = 0x18
    GETEUID = 0x19
    ACCESS = 0x21
    GETEGID = 0x2B
    GETTIMEOFDAY = 0x74
    SYSCTL = 0xCA
    SHM_OPEN = 0x10A
    SYSCTLBYNAME = 0x112
    GETTID = 0x11E
    ISSETUGID = 0x147
    STAT64 = 0x152
    LSTAT64 = 0x154
    GETENTROPY = 0x1F4

    MACH_ABSOLUTE_TIME_TRAP = 0xFFFFFFFD
    MACH_TIMEBASE_INFO_TRAP = 0xFFFFFFFFFFFFFFA7
    MACH_MSG_TRAP = 0xFFFFFFFFFFFFFFE1
    MACH_REPLY_PORT_TRAP = 0xFFFFFFFFFFFFFFE6

    KERNELRPC_MACH_VM_MAP_TRAP = 0xFFFFFFFFFFFFFFF1


class IosSyscallHandler(SyscallHandler):
    """System call handler for the iOS."""

    @classmethod
    def register(cls, emu):
        handlers = {
            IosSyscallNo.GETPID: cls.getpid,
            IosSyscallNo.GETUID: cls.getuid,
            IosSyscallNo.GETEUID: cls.geteuid,
            IosSyscallNo.ACCESS: cls.access,
            IosSyscallNo.GETEGID: cls.getegid,
            IosSyscallNo.GETTIMEOFDAY: cls.gettimeofday,
            IosSyscallNo.SYSCTL: cls.sysctl,
            IosSyscallNo.SHM_OPEN: cls.shm_open,
            IosSyscallNo.SYSCTLBYNAME: cls.sysctlbyname,
            IosSyscallNo.GETTID: cls.gettid,
            IosSyscallNo.ISSETUGID: cls.issetugid,
            IosSyscallNo.STAT64: cls.stat64,
            IosSyscallNo.LSTAT64: cls.lstat64,
            IosSyscallNo.GETENTROPY: cls.getentropy,
            IosSyscallNo.MACH_ABSOLUTE_TIME_TRAP: cls.mach_absolute_time,
            IosSyscallNo.MACH_TIMEBASE_INFO_TRAP: cls.mach_timebase_info_trap,
            IosSyscallNo.MACH_MSG_TRAP: cls.mach_msg_trap,
            IosSyscallNo.MACH_REPLY_PORT_TRAP: cls.mach_reply_port_trap,
            IosSyscallNo.KERNELRPC_MACH_VM_MAP_TRAP: cls.kernelrpc_mach_vm_map_trap,
        }

        emu.syscall_handlers.update(handlers)

    @staticmethod
    def getpid(emu):
        emu.set_retval(os.getpid())

    @staticmethod
    def getuid(emu):
        emu.set_retval(1)

    @staticmethod
    def geteuid(emu):
        emu.set_retval(1)

    @staticmethod
    def access(emu):
        # path = emu.read_string(emu.get_arg(0))
        emu.set_retval(0)

    @staticmethod
    def getegid(emu):
        emu.set_retval(1)

    @staticmethod
    def gettimeofday(emu):
        tv = emu.get_arg(0)

        emu.write_u64(tv, int(time.time()))
        emu.write_u64(tv + 8, 0)

        emu.set_retval(0)

    @staticmethod
    def sysctl(emu):
        emu.set_retval(0)

    @staticmethod
    def shm_open(emu):
        emu.set_retval(0x80000000)

    @staticmethod
    def sysctlbyname(emu):
        name = emu.read_string(emu.get_arg(0))
        oldp = emu.get_arg(2)
        oldlenp = emu.get_arg(3)

        if not oldp or not oldlenp:
            emu.set_retval(0)
            return

        if name == "kern.boottime":
            emu.write_u64(oldp, int(time.time()) - 3600 * 24)
            emu.write_u64(oldp + 8, 0)

        elif name == "kern.osvariant_status":
            # internal_release_type = 3
            emu.write_u64(oldp, 0x30)

        elif name == "hw.memsize":
            emu.write_u64(oldp, 4 * 1024 * 1024 * 1024)

        else:
            emu.logger.warning("Unhandled sysctlbyname: %s" % name)
            # raise RuntimeError("Unhandled sysctl command: %s" % name)

        emu.set_retval(0)

    @staticmethod
    def gettid(emu):
        tid = threading.current_thread().ident
        emu.set_retval(tid)

    @staticmethod
    def issetugid(emu):
        emu.set_retval(0)

    @staticmethod
    def stat64(emu):
        # path = emu.read_string(emu.get_arg(0))
        stat = emu.get_arg(1)

        # st_mode
        emu.write_u32(stat + 4, 0x4000)

        emu.set_retval(0)

    @staticmethod
    def lstat64(emu):
        # path = emu.read_string(emu.get_arg(0))
        stat = emu.get_arg(1)

        # st_mode
        emu.write_u32(stat + 4, 0x4000)

        emu.set_retval(0)

    @staticmethod
    def getentropy(emu):
        buffer = emu.get_arg(0)
        size = emu.get_arg(1)

        rand_bytes = bytes([random.randint(0, 255) for _ in range(size)])
        emu.write_bytes(buffer, rand_bytes)

        emu.set_retval(0)

    @staticmethod
    def mach_absolute_time(emu):
        emu.set_retval(int(time.time_ns() % (3600 * 10**9)))

    @staticmethod
    def mach_timebase_info_trap(emu):
        info = emu.get_arg(0)

        emu.write_u32(info, 1)
        emu.write_u32(info + 4, 1)

        emu.set_retval(0)

    @staticmethod
    def mach_msg_trap(emu):
        # msg = emu.get_arg(0)

        # msgh_size:
        # emu.write_u32(msg + 0x4, 36)

        # msgh_remote_port
        # emu.write_u32(msg + 0x8, 0)

        # msgh_id
        # emu.write_u32(msg + 0x14, 8100)

        emu.set_retval(6)

    @staticmethod
    def mach_reply_port_trap(emu):
        emu.set_retval(0)

    @staticmethod
    def kernelrpc_mach_vm_map_trap(emu):
        address = emu.get_arg(1)
        size = emu.get_arg(2)

        mem = emu.memory_manager.alloc(size)
        emu.write_pointer(address, mem)

        emu.set_retval(0)



import sys

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

# Endian type
EndianType = Literal["little", "big"]

BIG_ENDIAN: Literal["big"] = "big"
LITTLE_ENDIAN: Literal["little"] = "little"

# Supported arch
ARCH_ARM = 1
ARCH_ARM64 = 2

MODE_ARM = 1
MODE_THUMB = 2

# Supported os type
OS_ANDROID = 1
OS_IOS = 2

# The address of thread local storage (TLS)
TLS_ADDRESS = 0x10000
TLS_SIZE = 0x10000

# The address of stack
STACK_ADDRESS = 0x50000
STACK_SIZE = 0x50000

# The address of module
MODULE_ADDRESS = 0x10000000

# The address of heap
HEAP_ADDRESS = 0x8000000

# Minimum size of memory pool
MINIMUM_POOL_SIZE = 2**16



from .hooks import CStandardLibraryHooks
from .loader import ELFLoader

from chomper.types import BaseOs


class AndroidOs(BaseOs):
    """The Android environment."""

    def __init__(self, emu, **kwargs):
        super().__init__(emu, **kwargs)

        self.loader = ELFLoader(emu)

    def init_hooks(self):
        """Initialize the hooks."""
        CStandardLibraryHooks.register(self.emu)

    def initialize(self):
        """Initialize the environment."""
        self.init_hooks()
